/** @file cover-combine.cc
 * From a large set of clusters on a graph, choose a small set to cover
 * the graph itself, then combine the clusters to hit a target number.
 * The resulting clusters still overlap, but there are only a small
 * number of them now. 
 */

/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, University of British Columbia, 2010
 * 
 * History
 * -------
 * :2010-01-27: Initial coding 
 * :2010-01-28: Added core assignment 
*/

#define NOMINMAX
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <algorithm>
#include <vector>
#include <limits>
#include <set>
#include <map>

#include "sparfun.h"
#include "sparvec.h" // for tr1ns
#include "vecfun.h"
#include "sparfun_util.h"
#include "sparse_containers.hpp"

#include "set_cover.hpp"

void usage(FILE* s) {
  const char *usagestr =
    "cover-combine graphfile containfile maxvol outfile\n"
    "             [coretype] [beta-scale] [vol-penalty] [border-penalty]\n"
    "All options are positional.  To get the default value, specify _\n"
    "e.g. cover-combine graph.smat graph.escprobs graph.cores _ _ 10\n"
    "changes the vol-penalty to 10 without touching the beta-scale factor\n"
    "\n"
    "This procedure combines clusters and outputs an overlapping cluster\n"
    "matrix where the core for each vertex is denoted with a 1 for\n"
    "the value and each non-core vertex is denoted with a -1 for the\n"
    "the value.  The output is an smat file just like in the other cases.\n"
    "coretype determines how cores are assigned to clusters\n"
    "  best - agglomerates all the clusters, then assigns largest leavetime\n"
    "  each - each cluster picks its cores, and these are agglomerated\n"
    "\n"
    "defaults:\n"
    " coretype         : best\n"
    " beta-scale       : 1.1\n"
    " volume-penalty   : 5\n"
    " border-penalty   : 1";
  fprintf(s,"%s\n",usagestr);
  exit(-1);
}

double double_arg(const char *a,double d) {
  if (!a) { return d; }
  if (strcmp(a,"_")==0) {
    return d; 
  } else {
    return atof(a);
  }
}


/** Find the cluster with the best of the worse containments
 * That is, look at the vertex with worst containment in any cluster,
 * then find the cluster with the best of the worst.  This tells us
 * that a core had better be INSIDE the best of the worst containment.
 */
double compute_best_beta(sparserow* C)
{
  double maxval = std::numeric_limits<double>::max();
  std::vector<double> betas(C->n,maxval);
  for (int i=0; i<C->m; ++i) {
    for (int nzi=C->ai[i]; nzi<C->ai[i+1]; ++nzi) { 
      int v = C->aj[nzi];
      double contain = C->a[nzi]; // containment

      betas[C->aj[nzi]] = std::min(betas[C->aj[nzi]],contain);
    }
  }
  double bestbeta=std::numeric_limits<double>::min();
  for (int i=0; i<C->n; ++i) {
    if (betas[i]!=maxval) {
      bestbeta = std::max(betas[i],bestbeta);
    }
  }
  return bestbeta;
}

/** Assign vertices as cores of a cluster.
 * Given a compressed sparse row matrix where each row in a cluster
 * and the non-zeros in the row is the containment of the
 * node in the cluster, this script assigns cores if the containment
 * is higher than some threshold.
 * 
 * @param Cs the CSR matrix where each non-zero value is an escaping 
 *  probability
 * @param core_ind a vector mapping to the non-zero values in Cs, which
 *   is 1 if the vertex in the cluster is a core, or 0 otherwise.  This
 *   array has length nnz(Cs)
 * @param beta the value of beta to map a vertex to a core
 * @param core_assigned is an array to tell if a vertex already has
 *   a core assigned.  If so, then it does not get a new one.
 */
void assign_cores(const sparserow& Cs, int* core_ind, double beta, 
       int* core_assigned)
{
    for (size_t i=0; i<Cs.ai[Cs.m]; i++) {
        int v = Cs.aj[i];// get the vertex
        if (core_assigned[v]) { continue; } // don't mess with it
        if (Cs.a[i] > beta) {
            core_ind[i] = 1;
        }
    }
}

/** Validate a core assignment.
 * A valid core assignment should have every vertex assigned in at least
 * one core.  So an invalid core assignment has at least one vertex 
 * unassigned.  Thus, we just have to run through all the vertices
 * and make sure they are all associated with a core. 
 * 
 * This step is required so that the set_cover problem is well defined.
 * 
 * @param Cs the CSR matrix where the non-zeros state which vertices
 *  are assigned to a cluster
 * @param core_ind an array of length nnz(Cs) that states if a vertex
 *  is a core in that cluster or not.
 * @param core_assigned an array of length nverts determine which 
 *  vertices already have a core assigned.  This will be updated
 *  based on the assignments made in this round too...
 */
bool validate_cores(const sparserow& Cs, int* core_ind, int* core_assigned) {
    
    for (size_t i=0; i<Cs.ai[Cs.m]; i++) {
        if (core_ind[i]) {
            core_assigned[Cs.aj[i]] = 1;
        }
    }
    // make sure all vertices got set to be in a core
    for (size_t i=0; i<Cs.n; i++) {
        if (core_assigned[i] == 0) {
            return false;
        }
    }
    
    return true;
}

/** Fix a definite assignment from vectices to cores.
 * In a final overlapping clustering, each vertex has only one core.
 * This method produces such a mapping, where each vertex
 * is assigned to the cluster with largest containment.  This should
 * occur after a covering clustering is chosen.  At this stage, we know
 * there should be only one good cluster for a vertex, so we can 
 * safely and greedily assign.  
 * 
 * TODO Evaluate the output, it's possible that this would not
 * be ideal because cores should obey clusters.  Thus, it may
 * be better to use each cluster's cores directly and break ties
 * in that fashion instead.  
 * 
 * @param CCs a CSR matrix where each cluster is a row and
 *  the containment of the vertex is given by the value
 * @param core_map an array of length nverts (CCs->n) 
 *  that contains the cluster number of the core for each vertex
 */
void assign_best_cores(sparserow *CCs, int* core_map) {
    std::vector<double> best_contain_for_vertex(CCs->n,
        std::numeric_limits<double>::min());
    for (int i=0; i<CCs->m; ++i) {
        // run through all vertices in cluster i
        for (int nzi=CCs->ai[i]; nzi<CCs->ai[i+1]; ++nzi) {
            int j = CCs->aj[nzi];
            double contain = CCs->a[nzi];
            if (contain > best_contain_for_vertex[j]) {
                best_contain_for_vertex[j] = contain;
                core_map[j] = i;
            }
        }
    }
}

/** Assign cores to clusters
 * Same type of assignment as the other function, but different
 * output.  This function works with the final set of clusters.
 */
void assign_best_cores(int nverts, 
    std::vector< std::pair<int,int> >& new_clusters,
    std::vector< double >& leavetimes, int* core_map) 
{    
    std::vector<double> best_contain_for_vertex(nverts,
        std::numeric_limits<double>::max());
    typedef std::vector< std::pair<int, int> >::const_iterator it_type;
    std::vector< double >::const_iterator val_it = leavetimes.begin();
    for (it_type it=new_clusters.begin(), itend=new_clusters.end();
         it!=itend;++it,++val_it) {
        int i = it->first;
        int j = it->second;
        double contain = *val_it;
        if (contain > best_contain_for_vertex[j]) {
            best_contain_for_vertex[j] = contain;
            core_map[j] = i;
        }
    }  
}

/* Simple definitions for getting an identity map.
 */
template <typename type>
struct identity_map {
    type& operator[](type& t) { return t; }
};
template <class type>
identity_map<type> make_identity_map() { 
    return identity_map<type>();
}

struct sparserow_degreeof_functor {
    sparserow& graph;
    sparserow_degreeof_functor(sparserow& G) : graph(G) {}
    double operator()(const int vert) { 
        return (double)sr_degree(&graph, vert);
    }
};

/** 
 * @param volume_penalty
 * @param boundary_penalty
 * @param cover must be initialized to an array with Cs.m elements
 *  (the total number of sets).  At the end, this array will contain
 *  the indices of all the valid sets.  
 */
size_t cover_cores(sparserow& G, sparserow& Cs, int* core_ind, 
   double volume_penalty, double boundary_penalty,
   size_t* cover) {
    // copy our core_vertices into a set_of_sets
    // we can also compute the cost of each set while we are
    // doing this operation
    std::vector<double> cost(Cs.n);
    std::vector< std::vector<int> > set_of_sets(Cs.m);
    for (size_t i=0; i<Cs.m; i++) {
        size_t nclusterverts = Cs.ai[i+1]-Cs.ai[i];
        size_t ncore=0;
        // count the number of core vertices in this set
        for (int nzi=Cs.ai[i]; nzi<Cs.ai[i+1]; ++nzi) { 
            if (core_ind[nzi]) {
                ncore++;
            }
        }
        set_of_sets[i].resize(ncore);
        ncore = 0;
        // load all the core vertices into the set_of_sets
        for (int nzi=Cs.ai[i]; nzi<Cs.ai[i+1]; ++nzi) { 
            if (core_ind[nzi]) {
                set_of_sets[i][ncore] = Cs.aj[nzi];
                ncore++;
            }
        }
        double volume = (double)sr_graph_volume(
                                    &G, &Cs.aj[Cs.ai[i]], nclusterverts);
        double cut = (double)sr_graph_cutsize(
                                    &G, &Cs.aj[Cs.ai[i]], nclusterverts);
        cost[i] = volume_penalty*volume + boundary_penalty*cut;
    }
    
    size_t nverts = Cs.n;

    size_t *end = 
        approx::greedy_set_cover(nverts, set_of_sets, cost, 
            make_identity_map<int>(), sparserow_degreeof_functor(G), 
            cover);
    size_t val = end - cover;
    
    // make sure it's a valid set cover
    if (approx::validate_set_cover(nverts, set_of_sets, 
                    make_identity_map<int>(), cover, end) == false) {
        printf("SET COVER FAILED!\n");
        // TODO debug set cover failure output
        exit(-1);
    }
    return val;
}


    
/** Combine covers
 * Consider all the available clusters and try to combine them until
 * we get to maxvol.
 * Let clusters_small be the new of clusters
 * Let clusters be the original set of clusters
 * Let T be the set vertices that we have assigned to new clusters
 * 
 * While there are clusters left in the original set (clusters is not empty)
 *   Create an empty new cluster C
 *   While there are clusters left in the original set ...
 *     if the new cluster is empty
 *       Pick a cluster D from what's left in the original set (D in clusters)
 *         The cluster D should minimize the intersection with T 
 *     else
 *       Pick a cluster D from what's left in the original set (D in clusters)
 *         The cluster D should maximize the intersection with C
 *       If D and C is too large, then break....
 *     Add D to C
 *     assign any node with a core in D to have a core in C
 *     Remove D from the set of original clusters
 *   After the break
 *     Add C to clusters_small
 * 
 * The expensive part of this algorithm is that we need to figure out
 * which clusters to add 
 * 
 * @param core_map an array of size nverts giving the best cluster
 *  for each vertex.
 *     
 * @return the number of new clusters
 */    
int cover_combine(const sparserow& Gs, sparserow *CCs, int max_volume,
        std::vector< std::pair<int, int> >& clusters,
        int* core_map, std::vector<double>* leave_times)
{
    size_t nverts = Gs.n;
    size_t nclusters = CCs->m;
    
    // build a map from vertices in G to clusters in CCs
    sparserow* vert2cluster = sf_transpose(CCs, false);
    if (!vert2cluster) {
        printf("Error, memory allocation failed for vertex cluster map\n");
        exit(-1);
    }
    
    // build a map from clusters to their core vertices
    sparserow* cluster2core = 
        sf_index_int_vector_elements(core_map, nverts, CCs->m);
    if (!cluster2core) {
        printf("Error, memory allocation failed for cluster core map\n");
        exit(-1);
    }
    
    // allocate T, the set of newly assigned vertices
    std::vector<int> assigned_verts(nverts,0); 
    
    int new_cluster_num = 0; // how many clusters we've allocated
    
    // allocate C intersect T
    std::vector<int> cluster_overlap(nclusters,0);
    // this map isn't strictly necessary, but it simplifies version 1
    std::set<int> clusters_remaining;
    typedef std::set<int>::iterator clusters_remaining_iterator;
    std::vector<int> cluster_remaining_flag(nclusters,1); // for O(1) check
    for (size_t i=0; i<nclusters; i++) {
        clusters_remaining.insert(i); 
    }
    
    while (clusters_remaining.size()>0) {
        
        // allocate the new cluster
        sparse_map_array<int,double> new_cluster(nverts);
        int new_cluster_volume = 0;
        
        tr1ns::map<size_t, int> new_cluster_overlap;
        typedef tr1ns::map<size_t, int>::iterator cluster_map_iterator;
        
        for (clusters_remaining_iterator iter = clusters_remaining.begin();
             iter != clusters_remaining.end(); ++iter) {
            new_cluster_overlap[*iter] = 0;
        }
        
        while (clusters_remaining.size()>0) {
            int Dind = -1;
            if (new_cluster.size() == 0) {
                // pick a cluster to minimize the overall intersection
                // this step needs a heap, but not for version 1
                
                // the current min vol intersect
                int min_vol_intersect = Gs.ai[nverts]; 
                
                for (size_t i=0; i<nclusters; i++) {
                    if (cluster_remaining_flag[i]) {
                        if (cluster_overlap[i] < min_vol_intersect) {
                            Dind = i;
                            min_vol_intersect = cluster_overlap[i];
                        }
                    }
                }
                // Dind is the index of the seed cluster
            } else {
                int max_cluster_intersect = -1;
                // pick a cluster to maximize the intersection with new_cluster
                for (cluster_map_iterator iter = new_cluster_overlap.begin();
                     iter != new_cluster_overlap.end(); ++iter) {
                    if (iter->second > max_cluster_intersect) {
                        max_cluster_intersect = iter->second;
                        Dind = iter->first;
                    }
                }
                
                // check if the combined cluster is too big
                int prior_cluster_volume = new_cluster_volume;
                for (int nzi=CCs->ai[Dind]; nzi<CCs->ai[Dind+1]; ++nzi) {
                    int j = CCs->aj[nzi]; // vertex j in the cluster
                    if (new_cluster.count(j) == 0) {
                        // totally new vertex ...
                        prior_cluster_volume += sr_degree(&Gs,j);
                    }
                } 
                if (prior_cluster_volume > max_volume) {
                    // stop combining clusters, we've gotten too big
                    break;
                }
                // we actually update the volume later
            }
            
            // we should have a cluster Dind now
            assert(Dind >= 0);
            assert(cluster_remaining_flag[Dind] == 1);
            
            // remove D
            cluster_remaining_flag[Dind] = 0;
            clusters_remaining.erase(Dind);
            new_cluster_overlap.erase(Dind);
            
            // we now add all the vertices in D to the cluster
            for (int nzi=CCs->ai[Dind]; nzi<CCs->ai[Dind+1]; ++nzi) {
                int j = CCs->aj[nzi];
                if (new_cluster.count(j) == 0) {
                    // this vertex is not in the cluster
                    new_cluster.insert(j, CCs->a[nzi]);
                    int j_degree = sr_degree(&Gs, j);
                    new_cluster_volume += j_degree;
                    
                    // this means we have to go through and update 
                    // everything!  Blast!
                    
                    // if it has never even been assigned, do
                    // the full update against all remaining
                    // clusters
                    bool never_assigned = assigned_verts[j] == 0;
                    
                    for (int nzc=vert2cluster->ai[j]; 
                         nzc<vert2cluster->ai[j+1]; ++nzc) {
                        int ci = vert2cluster->aj[nzc];
                        if (ci==Dind) { continue; } // skip ourselves
                        assert( ci < nclusters ) ;
                        if (cluster_remaining_flag[ci]) {
                            
                            if (never_assigned) {
                                // add the volume of this vertex to
                                // the intersection of the cluster with T
                                cluster_overlap[ci] += j_degree;
                            }
                            
                            // TODO check if we can optimize
                            // this step based on 
                            new_cluster_overlap[ci] += j_degree;
                        }
                    }

                    if (never_assigned) {
                        // now actually assign the vertex
                        assigned_verts[j] = 1;
                    }
                } else {
                    // the vertex ix already in the cluster, but
                    // check if it's better contained in the new cluster
                    if (CCs->a[nzi] > new_cluster.get(j)) {
                        new_cluster.set(j, CCs->a[nzi]);
                    }
                }
                    
            }
            
            // update the core_map for all the vertex in D
            for (int nzi=cluster2core->ai[Dind]; 
                 nzi<cluster2core->ai[Dind+1]; ++nzi) {
                core_map[cluster2core->aj[nzi]] = new_cluster_num;
            }
        }
        
        // output the cluster
        for (int i=0; i<(int)new_cluster.size(); i++) {
            clusters.push_back(
                std::make_pair(new_cluster_num, new_cluster.elements[i]) );
            if (leave_times) {
                leave_times->push_back(new_cluster.values[i]);
            }
        }
        new_cluster_num++;
        
            
    }
    
    // free the map (TODO check leaks here with the actual structure)
    free_sparse(vert2cluster);
    free_sparse(cluster2core);
    
    return new_cluster_num;
}        

int main(int argc, char **argv) 
{
  using namespace std;
  double st;

  if (argc<5) { usage(stdout); } else if (argc>9) { usage(stderr); }
  const char *graphfilename = argv[1];
  const char *escprobfilename = argv[2];
  double max_volume_arg = double_arg(argv[3],-1.);
  const char *outfilename = argv[4];
  const char *coretype = argc>5?argv[5]:"best";
  double beta_scale=double_arg(argc>6?argv[6]:NULL,1.1);
  double volume_penalty=double_arg(argc>6?argv[7]:NULL,5.);
  double border_penalty=double_arg(argc>8?argv[8]:NULL,1.);
  
  if (strcmp(coretype, "best")!=0 && strcmp(coretype, "each")) {
      cerr << "invalid coretype: " << coretype << endl;
      exit(-1);
  }
  
  printf("Parameters\n");
  printf("----------\n");
  printf("%25s : %s\n", "graphfilename", graphfilename);
  printf("%25s : %s\n", "containmentfilename", escprobfilename);
  printf("%25s : %s\n", "maxvol", argv[3]);
  printf("%25s : %s\n", "outfilename", outfilename);
  printf("%25s : %s\n", "coretype", coretype);
  printf("%25s : %f\n", "beta_scale", beta_scale);
  printf("%25s : %i\n", "volume_penalty", (int)volume_penalty);
  printf("%25s : %i\n", "border_penalty", (int)border_penalty);
  printf("\n");


  sparserow Gs={0};
  sparserow Cs={0};
  int rval = 0;
  
  st = sf_time();
  rval |= load_generic(graphfilename,&Gs,false);
  if (rval) {
    cerr << "*** Error: couldn't open " << graphfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << graphfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    nodes = " << Gs.m << endl;
  cout << "    edges = " << sr_nedges(&Gs) << endl;
  cout << endl;
  
  st = sf_time();
  rval |= load_smat(escprobfilename,&Cs,true);
  if (rval) {
    cerr << "*** Error: couldn't open " << escprobfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << escprobfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    clusters = " << Cs.m << endl;
  cout << "    assignments = " << sr_nedges(&Cs) << endl;
  cout << endl;
  
  int nverts = std::max(Gs.m,Cs.m);
  
  // set the actual max_volume
  int max_volume;
  if (max_volume_arg < 0) {
    max_volume = std::max((int)(sr_nedges(&Gs)*0.01),100);
  } else if (max_volume_arg == 0) {
    printf("error invalid max_volume\n");
    exit(-10);
  } else if (max_volume_arg > 0 && max_volume_arg <= 1) {
    max_volume = (int)sr_nedges(&Gs)*max_volume_arg;
  } else {
    max_volume = (int)max_volume_arg;
  }
  printf("Using max_volume = %i\n", max_volume);
      
      
      
  //
  // determine the value type
  // if all the values in Es are larger than 1, then we assume
  // that they are exit-times, not escape probabilities
  // if all the values in Es are less than 1, then we assume they
  // are escape probabilities.  
  //
  

  // 
  // compute best beta
  //
  double best_beta = compute_best_beta(&Cs);
  printf("Found best-of-worse containment = %f\n",best_beta);

  //
  // compute cores
  //
  // this step is now done when we load them and
  // actually compute statistics of the final clustering
  //
  // we now assign cores in a slightly different way.  
  // first, we fix a beta_scale > 1, then look for all vertices
  // that will get a core under that assignment.  For all
  // vertices without a core, then we reduce beta until we assign
  // them, but there is a fixed level at which we assign cores, 
  // so that vertices with a good core assignment will get it.
  //
  std::vector<int> cores(Cs.ai[Cs.m],0);
  {
    std::vector<int> core_assigned(Gs.n,0);
    while (1) {
      printf("assigning cores with containment > %f\n", best_beta*beta_scale);
      assign_cores(Cs, &cores[0], best_beta*beta_scale, &core_assigned[0]);
      if (validate_cores(Cs, &cores[0], &core_assigned[0])) {
          break;
      } else {
          printf("invalid core set with beta_scale=%f ... repeating\n",
              beta_scale);
          beta_scale *= 0.9;
      }
    }
  }
  
  //
  // solve the set cover problem
  //
  sparserow* CCs = NULL;
  {
    std::vector<size_t> cover_sets(Cs.n,Cs.m+1);
    size_t nsets = cover_cores(Gs, Cs, &cores[0], 
            volume_penalty, border_penalty, &cover_sets[0]);
        
    printf("picked %Zi sets\n",nsets);
  
    // copy the sets to a new matrix
    CCs = sf_row_subset(&Cs, &cover_sets[0], nsets);
    free_sparse(&Cs); // free the old matrix
  }
  
  // 
  // build a mapping from vertices to cores
  //
  std::vector<int> core_map(Gs.n,-1);
  assign_best_cores(CCs,&core_map[0]);
  
  // 
  // combine clusters
  //
  
  std::vector< double > new_containment;
  std::vector< double >* containment_ptr = NULL;
  if (strcmp(coretype,"best")==0) {
      containment_ptr = &new_containment;
  }
  std::vector< std::pair<int, int> > new_clusters;
  int nclusters = cover_combine(Gs, CCs, max_volume, 
    new_clusters, &core_map[0], containment_ptr);
  printf("found %i clusters\n",nclusters);
  
  if (nclusters <= 0) {
    cout << "    Warning: no clusters generated" << endl;
  }
  
  if (containment_ptr) {
      // assign best cores again, but we need a different function
      assign_best_cores(Gs.n, new_clusters, *containment_ptr, &core_map[0]);
  }

  
  FILE *outfile = fopen(outfilename,"wt");
  if (outfile) {
    fprintf(outfile,"%u %u %Zu\n",nclusters,Gs.m,new_clusters.size());
    typedef std::vector< std::pair<int, int> >::const_iterator it_type;
    for (it_type it=new_clusters.begin(),itend=new_clusters.end();it!=itend;++it) {
      int core_value = -1;
      if (core_map[it->second]==it->first) {
          // this cluster is the core for this vertex, so output it
          core_value = 1;
      }
          
      fprintf(outfile,"%u %u %i\n",it->first,it->second,core_value);
    }
  } else {
    cerr << "*** Error: Couldn't write output to " << outfilename << endl;
    return (-1);
  }
  
  return 0;
}
