#!/usr/bin/env python

"""
evaluate-cover.py
=================

Evaluate a covering and the result of running a series of swaps.
For a graph, a set of clusters, and a result of a swap experiment, report:

1. The total volume of a cluster
   The maxiumum volume of a cluster
   The average volume of a cluster
   
2. The total cutsize of the clusters
   The maximum cutsize
   The average cutsize
   
3. The maximum overlap of any vertex
   The average overlap of a vertex

and if they input a set of swaps...
   
4. An estimate of rho, the swapping probability

and if they input a PageRank history

5. The total PageRank communication volume

History
-------
:2010-02-04: Initial version
"""

__author__ = 'David F. Gleich'

import sys

def usage():
    return """
    Usage: python evaluate-cover.py graph.smat graph.cores [graph.swaps] [graph.prhist]
    """

class FastSet:
    """ A set over a finite universe of items. """
    def __init__(self, set, n):
        self.n = n
        self.member = [ False for _ in xrange(n) ]
        for i in set:
            self.member = True
    def __contains__(self,item):
        return self.member[item]
    
        
    
class Graph:
    def __init__(self, n, adj):
        self.adj = adj
        self.n = n
        nnz = 0
        for i in xrange(0,n):
            nnz += len(adj[i])
        self.nnz = nnz
        
    def volume(self,S):
        vol = 0
        for s in S:
            vol += len(self.adj[s])
        return vol
    
    def cutsize(self,S):
        cut = 0
        # index the set
        curset = set(S)
        
        for s in S:
            for j in self.adj[s]:
                if j not in curset:
                    cut += 1
        return cut
        

def read_smat(filename):
    file = open(filename,'rt')
    header = file.readline()
    parts = header.split()
    m = int(parts[0])
    n = int(parts[1])
    nz = int(parts[2])
    adj = [ [] for _ in xrange(0,m) ]
    for line in file:
        parts = line.split()
        i = int(parts[0])
        j = int(parts[1])
        adj[i].append(j)
    return Graph(m, adj)
    
def average(seq):
    mean = 0
    n = float(len(seq))
    for i in seq:
        mean += float(i)/n
    return mean
    
class Clusters:
    def __init__(self, m, vertices, cores):
        self.clusters = vertices
        self.cores = cores
        self.nclusters = len(vertices)
        
        if self.nclusters != m:
            sys.stderr.write('Error wrong number of clusters\n')
            sys.exit(-1)
            
        maxvert = 0
        for i in xrange(self.nclusters):
            for j in self.clusters[i]:
                maxvert = max(maxvert, j)
        
        self.nverts=maxvert+1
        
    def report(self, graph, print_labels = True):
        if self.nverts != graph.n:
            sys.stderr.write('Error, wrong number of vertices %i != %i\n',
                self.nverts, graph.n)
            sys.exit(-1)
        
        volumes=[ 0 for _ in xrange(self.nclusters)]
        cutsizes=[ 0 for _ in xrange(self.nclusters)]
        
        for ci in xrange(self.nclusters):
            volumes[ci] = graph.volume(self.clusters[ci])
            cutsizes[ci] = graph.cutsize(self.clusters[ci])
        
        overlaps = [ 0 for _ in xrange(self.nverts) ]
        for ci in xrange(self.nclusters):
            for j in self.clusters[ci]:
                overlaps[j] += 1
                
        if print_labels: sys.stdout.write('%25s '%('clusters'))
        print "%i"%(self.nclusters)                
        
        if print_labels: sys.stdout.write('%25s '%('total volume'))
        print "%i"%(sum(volumes))
        if print_labels: sys.stdout.write('%25s '%('volume ratio'))
        print "%f"%(float(sum(volumes))/float(graph.nnz))
        if print_labels: sys.stdout.write('%25s '%('maximum volume'))
        print "%i"%(max(volumes))
        if print_labels: sys.stdout.write('%25s '%('average volume'))
        print "%f"%(average(volumes))
        
        if print_labels: sys.stdout.write('%25s '%('total cutsize'))
        print "%i"%(sum(cutsizes))
        if print_labels: sys.stdout.write('%25s '%('maximum cutsize'))
        print "%i"%(max(cutsizes))
        if print_labels: sys.stdout.write('%25s '%('average cutsize'))
        print "%f"%(average(cutsizes))
        
        if print_labels: sys.stdout.write('%25s '%('total overlap'))
        print "%i"%(sum(overlaps))
        if print_labels: sys.stdout.write('%25s '%('maximum overlap'))
        print "%i"%(max(overlaps))
        if print_labels: sys.stdout.write('%25s '%('average overlap'))
        print "%f"%(average(overlaps))
        
        
    
    
        
def read_clusters(filename):
    file = open(filename,'rt') 
    header = file.readline()
    parts = header.split()
    nclusters = int(parts[0])
    nverts = int(parts[2])
    clusters = [ [] for _ in xrange(nclusters) ]
    cores = [ -1 for _ in xrange(nverts) ]
    for line in file:
        parts = line.split()
        ci = int(parts[0]) # cluster
        vi = int(parts[1]) # vertex
        cv = int(parts[2]) # core vertex indicator cv=1 if its a core
        if cv == 1:
            if cores[vi] != -1:
                sys.stderr.write(
                    'Duplicate core detected: (%i,%i) and (%i,%i)\n'%(
                        ci,vi,cores[vi],vi))
                sys.exit(-1)
            cores[vi] = ci
        clusters[ci].append(vi)
    return Clusters(nclusters, clusters, cores)
    
                    
def estimate_swapping_from_data(filename):
    """
    The way these histograms are stored is that the 
    steps is recorded in the maximum swaps possible, it should
    (often) have a count of 0, but it's still there.
    """
    file = open(filename,'rt')
    trials = 0
    steps = 0
    histogram = {}
    for line in file:
        parts = line.split()
        swaps = int(parts[0])
        count = int(parts[1])
        histogram[swaps] = count
        trials += count
        steps = max(steps,swaps)
        
    rho = float(0)
    for (swaps,count) in histogram.items():
        rho += float(swaps)/float(steps) * float(count)/float(trials)
    return rho
    
def pagerank_communication_from_data(filename):
    """
    the contents of the file are 
    <iteration>  <computation> <communication> <residual>  
    """
    file = open(filename,'rt')
    communication = 0.
    for line in file:
        parts = line.split()
        communication += float(parts[2])
        
    return communication
    

def main():
    if len(sys.argv) < 3 or len(sys.argv) > 5:
        print usage()
        sys.exit(-1)
    
    graphfilename = sys.argv[1]
    coresfilename = sys.argv[2]
    
    print_label = True

    graph = read_smat(graphfilename)
    clusters = read_clusters(coresfilename)
    
    clusters.report(graph, print_label)
    
    if len(sys.argv)>3:
        rho = estimate_swapping_from_data(sys.argv[3])
        if print_label: sys.stdout.write('%25s '%('swapping prob'))
        print "%f"%(rho)

    if len(sys.argv)>4:
        prcomm = pagerank_communication_from_data(sys.argv[4])
        if print_label: sys.stdout.write('%25s '%('pagerank communication'))
        print "%g"%(prcomm)
        

if __name__=='__main__':
    main()
