%% Show overlapping clusters of a butterfly.

%% Setup paths 
addpath('../../matlab');
addpath('../tools');

%% Plot options
graphlw = 2; vtxsize=25; ptssize=100;
bkgcolor = 0.83*[1,1,1];
graphclr = 0.55*[1,1,1];
redclr = [0.8,0,0];

%% Load the graph
A = readSMAT('../../data/overlapping-clusters.smat');
xy = load('../../data/overlapping-clusters.xy');
xy(:,2) = -xy(:,2);

%%
% plot it
clf; hold on; axis off; set(gcf,'Color',bkgcolor);
[lx,ly]=gplot(A,xy); h=plot(lx,ly,'k-','LineWidth',graphlw);
set(h,'Color',graphclr); plot(xy(:,1),xy(:,2),'.','MarkerSize',vtxsize,'Color',graphclr); 
colormap(hot); axis square;
title('The graph','FontSize',32);
pause;

%% Partition the graph
% We want each cluster to have about 25% of the graph
fprintf('Computing clusters and containment scores ...\n');
pause;
maxvol=0.5;
H = hypercluster(A,'maxvol',0.5,'expand',nnz(A)*maxvol/25);

%% Compute the containment
C = containment(A,H,'alpha',0.99,'method','leavetime','maxsteps',150);
    
%% 
% Visualize the containment of each cluster
for hi=1:size(C,1)
    clf; hold on; axis off; set(gcf,'Color',bkgcolor);
    [lx,ly]=gplot(A,xy); h=plot(lx,ly,'k-','LineWidth',graphlw);
    set(h,'Color',graphclr); plot(xy(:,1),xy(:,2),'.','MarkerSize',vtxsize,'Color',graphclr); 
    colormap(hot); axis square;

    [i,hverts,probs] = find(C(hi,:));
    xys = xy(hverts,:);
    scatter(xys(:,1),xys(:,2),ptssize,probs,'filled')
    title({sprintf('Cluster %i and Containment', hi), '(white=good, black=bad)'},'FontSize',32);
    pause;
end

%% Combine clusters
fprintf('About to combine the clusters ... \n');
pause
S = combine_clusters(A,C,'maxvol',maxvol,'coretype','best');

%%
% Plot the cores
clf; hold on; axis off; set(gcf,'Color',bkgcolor);
[lx,ly]=gplot(A,xy); h=plot(lx,ly,'k-','LineWidth',graphlw);
set(h,'Color',graphclr); plot(xy(:,1),xy(:,2),'.','MarkerSize',vtxsize,'Color',graphclr); 
colormap(hot); axis square;

[cs vs ignore] = find(S==1);
scatter(xy(vs,1),xy(vs,2),ptssize,cs,'Filled');
colormap(hot);
title('Home clusters of vertices (color=cluster)','FontSize',32);
pause

%% 
%Show individual clusters
for ci=1:size(S,1);
    clf; hold on; axis off; set(gcf,'Color',bkgcolor);
    [lx,ly]=gplot(A,xy); h=plot(lx,ly,'k-','LineWidth',graphlw);
    set(h,'Color',graphclr); plot(xy(:,1),xy(:,2),'.','MarkerSize',vtxsize,'Color',graphclr); 
    colormap([0 0 0; graphclr; redclr]); axis square;
    set(gca,'Clim',[-1,1]);
    cv = S(ci,:);
    scatter(xy(:,1),xy(:,2),ptssize,cv,'Filled');
    title({sprintf('Final cluster %i', ci), '(red=home, black=extra)'},'FontSize',32);
    pause
end

%%
%Show overlap
clf; hold on; axis off; set(gcf,'Color',bkgcolor);
[lx,ly]=gplot(A,xy); h=plot(lx,ly,'k-','LineWidth',graphlw);
set(h,'Color',graphclr); plot(xy(:,1),xy(:,2),'.','MarkerSize',vtxsize,'Color',graphclr); 
colormap(hot); axis square;

olap = sum(spones(S));
scatter(xy(:,1),xy(:,2),ptssize,olap,'Filled');
title({'Overlap', '(red = 1, white = 2)'},'FontSize',32);
pause;

%% 
% evaluate
stats = overlapping_evaluation(A,S);
home;
fprintf('Volume ratio = %f\n', stats.vol_ratio);
fprintf('Swapping probability = %f\n', stats.rho);
fprintf('PageRank communication = %i\n', stats.pagerank_communication);
