/*
 *      set_cover.hpp
 *      
 *      Copyright 2009 David Gleich <dgleich@stanford.edu>
 *      
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

/**
 * @file set_cover.hpp
 * @author David F. Gleich
 * @brief Approximation algorithms for the set covering problem.
 */

#ifdef __GNUC__
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#elif defined(__MSVC__)
#include <unordered_map>
#include <unordered_set>
#endif

#include <assert.h>
#include <vector>
#include <limits>

/** A namespace for approximation algorithms */
namespace approx {
    
    namespace detail {
        /**
         * A simple updatable heap assuming that 
         * the items are integers, and the values 
         * are a simple comparable type.  This heap also
         * assumes that you know the maximum number of 
         * entries.
         */
        template <class ItemType, class ValueType>
        class simple_updatable_heap {
            const size_t invalid_pos; // must come before locations!
            std::vector<ItemType> heap;
            std::vector<ValueType> values;
            std::vector<size_t> locations;
            size_t heap_size;
            size_t max_items;
            
            
            public:
            /**
             * Initialize the heap, this requires the total number of
             * elements.
             * @param n the maximum number of items in the heap
             */
            simple_updatable_heap(size_t n)
                : invalid_pos(std::numeric_limits<size_t>::max()),
                  max_items(n), heap(n), locations(n, invalid_pos), values(n),
                  heap_size(0)
            {}
            
            /* Binary tree in array guide
             Node 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
             Par  -  0  0  1  1  2  2  3  3  4  4  5  5  6  6  7  7
              left child = 2*(j+1)-1 (j=6 =>  left = 13)
             right child = 2*(j+1)   (j=3 => right = 8)
                  parent = (j-1)/2
             */
          
            /** 
             * Move an element up the heap to its' correct position
             * @param j the location of the item to move
             */
            size_t heap_up(size_t j) {
                while (1) {
                    if (j==0) { break; } // the element is at the top
                    ItemType heapj = heap[j];
                    size_t j2 = (j-1)/2;
                    ItemType heapj2 = heap[j2];
                    if (values[heapj2] < values[heapj]) {
                        break; // the parent is smaller, so stop
                    } else {
                        // the parent is larger, so swap
                        heap[j2] = heapj; locations[heapj] = j2;
                        heap[j] = heapj2; locations[heapj2] = j;
                        j = j2;
                    }
                }
                return (j);
            }
            size_t heap_down(size_t k) {
                ItemType heapk = heap[k];
                while (1) {
                    size_t i=2*(k+1)-1;
                    if (i>=heap_size) { break; } // end of heap
                    if (i<heap_size) { 
                        // pick between children (unneeded if i==heap_size-1)
                        ItemType left=heap[i];
                        ItemType right=heap[i+1];
                        if (values[right]<values[left]) { 
                            i=i+1; // pick the smaller right child
                        }
                    }
                    if (values[heapk] < values[heap[i]]) {
                        // k is smaller than both children, so end
                        break;
                    } else {
                        heap[k] = heap[i]; locations[heap[i]]=k;
                        heap[i] = heapk; locations[heapk] = i;
                        k=i;
                    }
                }
                return (k);
            }                
            
            /** Return the number of items in the heap */
            size_t size() { return heap_size; }
            /** Return the capacity of the heap */
            size_t capacity() { return max_items; }
          
            /**
             * Insert an element
             */
            void insert(ItemType item, ValueType val) {
                if (locations[item] != invalid_pos) { update(item,val); }
                heap[heap_size] = item;
                values[item] = val;
                locations[item] = heap_size;
                heap_size++;
                heap_up(heap_size-1);
            }
            
            /** Peek at the smallest value */
            ValueType minvalue() {
                assert(heap_size > 0);
                return values[heap[0]];
            }
            ItemType min() {
                assert(heap_size >0);
                return heap[0];
            }
            /**
             * Pop an element
             */
            void pop() {
                if (heap_size==0) { return; }
                ItemType v = heap[0];
                locations[v] = invalid_pos;
                heap[0] = heap[heap_size-1]; // promote the last element
                locations[heap[0]] = 0;
                heap_size--;
                heap_down(0); // move the first element down
            }
            /** Change the value of an element */
            void update(ItemType item, ValueType val) {
                if (locations[item] == invalid_pos) { return; }
                values[item] = val;
                // move the element down, then possibly back up
                // I believe this can be optimized by checking
                // if val < items[val] -> heap_up
                // if val > items[val] -> heap_down
                // but this is not really any worse
                heap_up(heap_down(locations[item]));
            }
        };
    }
    
    /** Check if a set cover is valid 
     * @param n the size of the universe of items (maximum index)
     * @param set_of_sets valid sets
     * @param item_index a map between set entries and linear (0:n-1)
     *   incides
     * @param start the start of the Output Iterator from greedy_set_cover
     * @param end the end of the OutputIterator from greedy_set_cover
     * @return true if the sets cover all items, false otherwise
     */
    template < typename SizeType, typename SetOfSets, 
        typename ItemIndexMap, typename SetIterator >
    bool validate_set_cover(
            SizeType n,
            SetOfSets& set_of_sets,
            ItemIndexMap item_index,
            SetIterator start, SetIterator end)
    {
        std::vector<int> covered(n, 0);
        for (; start != end; ++start) {
            size_t i = *start;
            for (size_t j=0; j<set_of_sets[i].size(); j++) {
                covered[item_index[set_of_sets[i][j]]] = 1;
            }
        }
        for (SizeType i=0; i<n; i++) {
            if (covered[i]==0) {
                return false;
            }
            return true;
        }
    }
                
    
    /** Approximate the mininum cost set cover with a greedy algorithm.
     * 
     * Mathematically, the minimum cost set cover problem is
     *   min  sum c_i x_i
     *   s.t. union over non-zero x_i of S_i = U (universe of items)
     *        x_i is binary
     * 
     * @param n the number of items in the universe
     * @param set_of_sets a collection of itemsets that define the set 
     *  covering problem
     *      size_t set_of_sets.size()
     *      set_of_sets[i].size() for 0 < i < set_of_sets.size()
     *      set_of_sets[i][j] is an item 0 < j < set_of_sets[i].size()
     * @param cost the cost of a set
     *      cost[i] for 0 < i < set_of_sets.size()
     *      CostVector::value_type
     * @param item_index a mapping between the items and a linear (0-n-1) 
     *      item_index[set_of_sets[i][j]] must be in [0,n-1] for all valid i, j
     * @param item_value_func a functor to compute the cost of an item 
     * 
     * @example 
     *  size_t n=5;
     *  std::vector<double> cost(3);
     *  std::vector< std::vector< int > > set_of_sets(3);
     *  set_of_sets[0].push_back(0,1,2)
     *  set_of_sets[1].push_back(1,2,3)
     *  set_of_sets[2].push_back(0,4)
     *  cost[0] = 1.0; cost[1]
     *  greedy_set_cover
     */

    
    template < typename SizeType, typename SetOfSets, class CostVector,
        typename ItemIndexMap, typename ItemValueFunc, typename OutputIterator>
    OutputIterator greedy_set_cover(
            SizeType n,
            SetOfSets& set_of_sets,
            CostVector cost,
            ItemIndexMap item_index,
            ItemValueFunc item_value_func,
            OutputIterator picked
            )
    {
        typedef typename CostVector::value_type numtype;
        // TODO typedef ItemIndexMap::value_type indtype;
        
        numtype infval = std::numeric_limits<numtype>::max();
        
        // first, compute the value for each set
        size_t nsets = set_of_sets.size();
        std::vector<numtype> set_value(nsets);
        for (size_t i=0; i<nsets; i++) {
            numtype value = 0;
            for (size_t j=0; j<set_of_sets[i].size(); j++) {
                value += item_value_func(set_of_sets[i][j]);
            }
            set_value[i] = value;
        }
        
        // second, compute a map between items and sets
        std::tr1::unordered_map<SizeType, std::vector<size_t> > item_in_sets;
        for (size_t i=0; i<nsets; i++) {
            for (size_t j=0; j<set_of_sets[i].size(); j++) {
                item_in_sets[item_index[set_of_sets[i][j]]].push_back(i);
            }
        }
        
        // third, instantiate a mutable_queue with this data
        detail::simple_updatable_heap<size_t, numtype> heap(nsets);
        std::tr1::unordered_set<size_t> uncovered;
        for (size_t i=0; i<nsets; i++) {
            for (size_t j=0; j<set_of_sets[i].size(); j++) {
                uncovered.insert(item_index[set_of_sets[i][j]]);
            }
            numtype value;
            if (set_value[i]>0) {
                value = cost[i]/set_value[i];
            } else {
                value = infval;
            }
            heap.insert(i, value);
        }
            
        while (uncovered.size() > 0) {
            size_t i = heap.min();
            numtype setval = heap.minvalue();
            heap.pop();
            
            // add i to the list of sets!
            *picked++ = i;
            
            for (size_t j=0; j<set_of_sets[i].size(); j++) {
                size_t item = item_index[set_of_sets[i][j]];
                if (uncovered.count(item) > 0) {
                    uncovered.erase(item);
                    for (size_t k=0; k<item_in_sets[item].size(); k++) {
                        // item is also in set k, so that set isn't
                        // as valuable anymore
                        set_value[k] -= item_value_func(set_of_sets[i][j]);
                        numtype value;
                        if (set_value[k]>0) {
                            value = cost[k]/set_value[k];
                        } else {
                            value = infval;
                        }
                        heap.update(k, value);
                    }
                }
            }
        }
        
        return picked;
    } 
    
}; // approx namespace
