function stats = overlapping_evaluation(A,S,ntrials)
% OVERLAPPING_EVALUATION Report statistics on an overlapping clustering
%
% stats = overlapping_evaluation(A,S) reports
%   rho: an estimate of the swapping probability
%   average_overlap: average overlap of vertices
%   overlap_ratio: sum of vertices in each cluster/vertices of A
%   max_overlap: maximum overlap of any vertex
%   average_vol: average volume of any cluster
%   maximum_vol: maximum volume of any cluster
%   minimum_vol: minimum volume of any cluster
%   total_vol: sum of degrees for each vertex in each cluster
%   vol_ratio: total_vol/vol(G)
%
% ... = overlapping_evaluation(A,S,ntrials) increases the number of trials
% in the random walk evaluation of rho
%

% History
% -------
% :2010-02-03: Initial version

if ~exist('ntrials','var') || isempty(ntrials), ntrials=500; end

T=10000;
[total,swaps,count] = count_swaps(A,S,'ntrials',ntrials,'nsteps',T);
[communication,computation] = pagerank_communication(A,S);
stats.nverts = size(A,1);
stats.nedges = nnz(A);
stats.rho  = sum((swaps./T).*count/sum(count));
stats.pagerank_communication = communication;
stats.pagerank_computation = computation;
overlap = full(sum(spones(S)));
stats.overlap_ratio = sum(overlap)/nnz(A);
stats.max_overlap = max(overlap);
stats.average_overlap = mean(overlap);
degrees = full(sum(A,2));
[si,sj] = find(S);
cluster_volumes = full(accumarray(si,degrees(sj)));
stats.total_vol = sum(cluster_volumes);
stats.vol_ratio = sum(cluster_volumes)/nnz(A);
stats.average_vol = mean(cluster_volumes);
stats.maximum_vol = max(cluster_volumes);
stats.minimum_vol = min(cluster_volumes);

