function [communication,computation,hist,x]=pagerank_communication(A,S,varargin)
% PAGERANK_COMMUNICATION Count communication operations in distributed
% PageRank in an overlapping clustering.
%
% This command is not thread safe, it always writes to a file 
% in the directory with the command.
%
% [total_communication,total_computation,history,x] =
%    pagerank_communication(A,S,alpha,tol,maxit) returns the total
% amount of communication, the total computation steps,  the 
% evolution of both as a function of the itertion, i.e.
%   history(i) = [iter, computation, communication, residual] 
% and finally the pagerank vector x itself.  
%
% If S==1 (exactly) then it writes out information for a single cluster
% containing the entire graph.
%
% ... = pagerank_communication(A,S,'key',value, ...)
% specify optional arguments:
%   'alpha' = [float | {0.85}] : number of steps in the walk
%   'maxit' = [int | {10000}] : maximum number of iterations
%   'tol' = [float | {1}] : the tolerance of the computation.  This value
%   is in terms of amounts of 1/n where n is the size of the graph.  Thus,
%   1 is a reasonable tolerance to use for large graphs.  For small graphs,
%   0.1 may be superior.  

% David F. Gleich
% Copyright, 2010

% History
% -------
% :2010-10-07: Initial version
% :2010-10-16: Added option for single cluster

mydir = fileparts(mfilename('fullpath'));
% parse options
options = struct(varargin{:});
mycmd = fullfile(mydir,'..','overlap-clusters','bin','pagerank-communication');
mygraphfile = fullfile(mydir,'pagerank-communication.smat');
myclusterfile = fullfile(mydir,'pagerank-communication.cores');
myoutfile = fullfile(mydir,'pagerank-communication.hist');
myprfile = fullfile(mydir,'pagerank-communication.pr');

writeSMAT(mygraphfile,A,struct('format','%i'));
if S==1
    writeSMAT(myclusterfile,ones(1,size(A,1)),struct('format','%i'));
else
    writeSMAT(myclusterfile,S,struct('format','%i'));
end

alpha = 0.85;
tol = 1;
maxit = 10000;

if isfield(options,'alpha'), alpha = options.alpha; end
if isfield(options,'tol'), tol = options.tol; end
if isfield(options,'maxit'), maxit = options.maxit; end

mycmd = [mycmd ' ' mygraphfile ' ' myclusterfile ' ' myprfile ' ' myoutfile];
mycmd = [mycmd ' ' ...
    sprintf('%f %f %i',alpha,tol,maxit)];

status = system(mycmd);
if status==0
    hist = load(myoutfile);
    x = load(myprfile);
    communication = sum(hist(:,3));
    computation = sum(hist(:,2));
end