/** @file average-swaps.cc
 * Run a random walk in a set of overlapping clusters and report
 * statistics about how many times a vertex must swap between clusters.
 */

/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, University of British Columbia, 2010
 * 
 * History
 * -------
 * :2010-02-01: Initial coding.  
 * :2010-02-04: Added output of total trials
 * :2010-02-05: Added random restarts
*/

#define NOMINMAX
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <algorithm>
#include <vector>
#include <limits>
#include <set>
#include <map>

#ifdef __APPLE__
#include <tr1/unordered_set>
#else 
#include <unordered_set>
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__APPLE__)
#define tr1ns std::tr1
#else
#define tr1ns std
#endif

#include "sparfun.h"
#include "sparvec.h" 
#include "vecfun.h"
#include "sparfun_util.h"

void usage(FILE* s) {
  const char *usagestr =
    "average-swaps graphfile corefile outfile\n"
    "             [steps] [ntrials] [seed]\n"
    "All options are positional.  To get the default value, specify _\n"
    "e.g. average-swaps graph.smat graph.cores swaps.hist _ 10\n"
    "changes ntrials to 10 without touching the steps factor\n"
    "\n"
    "Simulate a random walk on a set of overlapping clusters and\n"
    "count the number of swaps between clusters made while running\n"
    "\n"
    "defaults:\n"
    " steps     : 100\n"
    " ntrials   : 1000\n"
    " seed      : 0\n";
  fprintf(s,"%s\n",usagestr);
  exit(-1);
}

double double_arg(const char *a,double d) {
  if (!a) { return d; }
  if (strcmp(a,"_")==0) {
    return d; 
  } else {
    return atof(a);
  }
}

/** Simulate a random walk and return the number of cluster swaps
 * @param r the sparse row matrix for the graph
 * @param clusters any object with a .ismember(vertex,cluster) function
 * @param core_map any object with a [vertex] operator to map vertices to cores
 * @param v the starting vertex
 * @param nsteps the number of steps
 */
template <class OverlappingClusters, class CoreMap>
size_t particle_walk(sparserow *r, OverlappingClusters& clusters,
     CoreMap& core_map, int v, size_t nsteps)
{
    size_t nswaps = 0;
    int cluster = core_map[v];
    while (nsteps > 0) {
        int deg = sr_degree(r,v);
        if (deg > 0) {
            // pick a neighbor at random
            int randneighbor = sf_randint(0, deg-1);
            int u = r->aj[r->ai[v]+randneighbor];
            // check if it's in the same cluster
            if (clusters[cluster].count(u)>0) {
                // it's in the cluster
            } else {
                // it isn't in the cluster, swap
                cluster = core_map[u];
                nswaps ++;
            }
            v = u;
        } else {
            // randomly restart, but count it as a free swap
            v = sf_randint(0, r->n-1);
            cluster = core_map[v];
        }
        
        nsteps--;
    }
    return nswaps;
}

/** Simulate a set of particle walks and count statistics
 * @param r the graph for the walk
 * @param clusters a structure for an overlapping clustering
 */ 
template <class OverlappingClusters, class HistMapType, class CoreMap>
size_t count_swaps_in_particle_walk(sparserow *r, 
    OverlappingClusters& clusters,
    CoreMap& core_map,
    int nwalks, int nsteps,
    HistMapType& hist)
{
    int nverts = r->n;
    size_t total_swaps = 0;
    // for each walk
    for (int i=0; i<nwalks; i++) {
        // pick a random starting vertex
        int v = sf_randint(0,r->n-1);
        // run the walk
        size_t nswaps = particle_walk(r, clusters, core_map, v, nsteps);
        total_swaps += nswaps;
        hist[nswaps]++;
        printf("walk %5i : %7Zi swaps\n", i, nswaps);
    }
    return (total_swaps);
}    

/** Load an overlapping clustering into a vector of sets
 * This loads an smat file into a vector of sets, one set per
 * cluster
 * @param filename the name of the file
 * @param clusters an uninitialized structure for the clusters
 * @param core_map an uniniitalized structure for the core map from
 *   vertices to number of cores
 */
int load_overlapping_clusters(const char *filename,
    std::vector< tr1ns::unordered_set<int> >& clusters,
    std::vector<int>& core_map)
{
    FILE *f = fopen(filename, "rt");
    if (!f) {
        fprintf(stderr, "error opening %s\n", filename);
        return (-1);
    }

    int m, n, nz;
    if (fscanf(f, "%u %u %u", &m, &n, &nz)==3) {
        clusters.resize(m);
        core_map.resize(n);
        for (size_t ci=0; ci<n; ++ci) {
            core_map[ci] = -1;
        }
        int i, j, nzi = nz;
        double a;
        while (nzi-- > 0) {
            if (fscanf(f, "%u %u %lf", &i, &j, &a) != 3) {
                fprintf(stderr, "error reading %s\n", filename);
                return (-1);
            }
            if (i >= 0 && i < m) { 
                clusters[i].insert(j);
                if (a==1) {
                    if (core_map[j]!=-1) {
                        fprintf(stderr,
                            "duplicate core detected: line %i, vertex %u\n",
                            nz-nzi+1, j);
                        return (-3);
                    }
                    core_map[j]=i;
                }
            } else { 
                fprintf(stderr, 
                    "vertex-cluster out of range (0,%i): line %i - %u %u %lf\n",
                    m, nz-nzi+1, i, j, a);
                return (-2);
            }
        }
    } else {
        fprintf(stderr, "error reading %s\n", filename);
        return (-1);
    }
    
    // check the core_map
    for (size_t i=0; i<n; i++) {
        if (core_map[i]==-1) {
            fprintf(stderr,
                "invalid core_map for %s, vertex %Zi unassigned\n", 
                filename, i);
                return (-11);
        }
    }
    
    return 0;
}
    

int main(int argc, char **argv) 
{
  using namespace std;
  double st;

  if (argc<4) { usage(stdout); } else if (argc>7) { usage(stderr); }
  const char *graphfilename = argv[1];
  const char *corefilename = argv[2];
  const char *outhistfilename = argv[3];
  size_t steps=(size_t)double_arg(argc>4?argv[4]:NULL,100.);
  size_t ntrials=(size_t)double_arg(argc>5?argv[5]:NULL,100.);
  double dseed = double_arg(argc>6?argv[6]:NULL,0.);
  
  // parse the seed if its negative
  unsigned long useed = 0;
  if (dseed < 0) {
      useed = (unsigned long)sf_time();
  } else {
      useed = (unsigned long)dseed;
  }
      
  
  printf("Parameters\n");
  printf("----------\n");
  printf("%25s : %s\n", "graphfilename", graphfilename);
  printf("%25s : %s\n", "containmentfilename", corefilename);
  printf("%25s : %s\n", "outhistfilename", outhistfilename);
  printf("%25s : %Zi\n", "steps", steps);
  printf("%25s : %Zi\n", "ntrials", ntrials);
  printf("%25s : %Zi\n", "seed", useed);
  printf("\n");

  sf_srand(useed);

  sparserow Gs={0};
  sparserow Cs={0};
  int rval = 0;
  
  st = sf_time();
  rval |= load_generic(graphfilename,&Gs,false);
  if (rval) {
    cerr << "*** Error: couldn't open " << graphfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << graphfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    nodes = " << Gs.m << endl;
  cout << "    edges = " << sr_nedges(&Gs) << endl;
  cout << endl;
  
  st = sf_time();
  std::vector< tr1ns::unordered_set<int> > clusters;
  std::vector< int > core_map;
  rval |= load_overlapping_clusters(corefilename,clusters,core_map);
  if (rval) {
    cerr << "*** Error: couldn't open " << corefilename << endl;
    return (-10); 
  }
  cout << "Loaded " << corefilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    clusters = " << clusters.size() << endl;
  cout << endl;

  
  int nverts = Gs.n;
  assert(nverts == core_map.size());
  int nclusters = clusters.size();
  
    
 
  
  tr1ns::unordered_map< int, int > swaps_histogram;
  size_t total_swaps = count_swaps_in_particle_walk(&Gs, clusters, core_map, 
    steps, ntrials, swaps_histogram);
  
  printf("\n");
  printf("Total swaps: %Zi\n", total_swaps);
  printf("\n");
  
  
  // add in a ficticious swap at steps, to record the number of steps
  // note that steps is an upper-bound on the number of swaps, so 
  // recording this as such makes total sense.
  if (swaps_histogram.count(steps) == 0) {
      swaps_histogram[steps] = 0;
  }
  
  
  FILE *outfile = fopen(outhistfilename,"wt");
  if (outfile) {
    typedef tr1ns::unordered_map< int,int >::const_iterator it_type;
    
    for (it_type it=swaps_histogram.begin(),itend=swaps_histogram.end();
         it!=itend;++it) {
      fprintf(outfile,"%u %u\n",it->first, it->second);
    }
    
    fclose(outfile);
  } else {
    cerr << "*** Error: Couldn't write output to " << outhistfilename << endl;
    return (-1);
  }
  
  return 0;
}

