#!/usr/bin/env python

"""
cluster_graclus.py
=====================

Construct a partiting of a graph with graclus.  We need this code because
graclus likes to produce clusters with huge volumes.  It isn't fair 
to compare these against our algorithms.

History
-------
:2010-10-27: Initial version
"""

import sys
import os
import time
import random
from math import ceil, floor
from optparse import OptionParser

from overlap_help import *   

""" Return a parser for the command line options. """
def options_setup():
  usage = "usage: %prog [options] graphfilename"
  parser = OptionParser(usage=usage)
  parser.add_option("","--maxvol",dest="maxvol",metavar="int|float",
    help="the maximum ideal volume of any cluster expressed as " + 
    "a number of edges (int) or a fraction of total edges (float) " + 
    "the default is max(0.01*nedges,100)",
    default=None,type="float")
  parser.add_option("","--slack",dest="slack",metavar="float",
    help="the maximum violation in maxvol (as a fraction), "+
    "the default is 0.05, i.e. 5\% extra edges",
    default=0.05,type="float")
  parser.add_option("-o","--out",dest="outfilename",metavar="FILE",
    help = "the output filename for the hyperclustering, " + 
    "the default value is graphname.hcluster",default=None)
  parser.add_option("-t","--maxtrials",dest="ntrials",metavar="int",
    help = "the number of trials to take",default=10)
  parser.add_option("-v","--verbose",action="store_true",dest="verbose")
  return parser
  
def main():
  parser = options_setup()
  (options,args) = parser.parse_args()
  if len(args) != 1:
    parser.error("incorrect number of arguments")
  
  verbose = options.verbose
  
  graphfilename = args[0]
  
  if verbose:
    print "reading %s ...."%(graphfilename)
    
  graph = read_graph(graphfilename)
  nverts = graph.nverts()
  nedges = graph.nedges()
  
  if verbose:
    print "nverts=%i; nedges=%i"%(nverts,nedges)
    
  # find the path to graclus
  graclusexe = os.path.abspath( os.path.join(__file__,'..','..',
                'graclus1.2','graclus'))
                
  if verbose:
    print "graclusexe: %s"%(graclusexe)
  
  if not os.path.isfile(graclusexe):
    print "graclus command: %s does not exist or is not a file"%(graclusexe)
    sys.exit(1)
    
  if options.outfilename is None:
    (base,ext) = os.path.splitext(graphfilename)
    options.outfilename = base + ".hcluster"
    
  if options.maxvol is None:
    options.maxvol = max(ceil(0.01*nedges),100)
    
  elif options.maxvol < 1:
    options.maxvol = ceil(nedges*options.maxvol)
  
  nparts = ceil(nedges/float(options.maxvol))
  ntrials = options.ntrials
  
  # print the parameters
  print "    maxvol: %i"%(options.maxvol)
  print "     slack: %f"%(options.slack)
  print "   ntrials: %i"%(options.ntrials)
  print "    output: %s"%(options.outfilename)
  
  slack = 1 + options.slack
  maxvol = slack*options.maxvol
  
  
  assert(maxvol > 0)
  assert(options.ntrials > 0)
  
  clusters = []
  nclusters = 0
  
  np = nparts
  for run in xrange(options.ntrials):
    
    # run graclus
    gracluscmd = "%s %s %i"%(graclusexe, graphfilename, np)
    
    if verbose:
      print "running: %s"%(gracluscmd)
    
    rval,out,err = run_command(gracluscmd)
    
    if rval != 0:
      print "error running graclus with %s"%(gracluscmd)
      print "continuing ..."
      continue
    
    # graclus writes to the current directory
    [graphpath,localgraphname] = os.path.split(graphfilename)
    partvecname = "%s.part.%i"%(localgraphname,np)
    
    # read in the partition vector
    if verbose:
      print "reading %s"%(partvecname)
      
    partmap = {}
    partfile = open(partvecname,'rt')
    for i,v in enumerate(partfile):
      curcluster = int(v)
      if curcluster in partmap:
        mycluster = partmap[curcluster]
      else:
        mycluster = nclusters
        partmap[curcluster] = nclusters
        nclusters += 1
      clusters.append((i,mycluster))
    partfile.close()
    
    if verbose:
      print "removing %s"%(partvecname)
    os.remove(partvecname)
    
    cluster_maxvol = maximum_cluster_volume(graph,clusters)
    if cluster_maxvol <= maxvol*slack:
      break
    else:
      print ("  clusters violated maxvol (%i > %i) "%(cluster_maxvol,maxvol) +
        "repeating with more partitions ... ")
      np = np*2  
      clusters = []
      
  
  if cluster_maxvol > maxvol:
    print "** couldn't find a clustering to satisfy maxvol<=%i in %i trials\n"%(
      maxvol,ntrials)
    sys.exit(-1)
    
  # now we have to pack the clusters together.
  # the use of options.maxvol here is deliberate to prevent overpacking
  (nclusters, clusters) = pack_clusters(graph,clusters,options.maxvol)
  
  if verbose:
    print "writing %s"%(options.outfilename)
  # now write the hyper clustering
  hclustfile = open(options.outfilename,'wt')
  hclustfile.write('%i %i %i\n'%(nclusters, nverts, len(clusters)))
  for a in clusters:
    hclustfile.write('%i %i 1\n'%(a[1],a[0]))
  hclustfile.close()
  
  
  
if __name__=='__main__':
  main()
    
  
