#!/usr/bin/env python

"""
evaluate-clustering.py
=================

Evaluate a hypercluster and report for each cluster

the size 
the volume
the cutsize
the conductance

History
-------
:2010-10-25: Initial version
"""

__author__ = 'David F. Gleich'

import sys

def usage():
    return """
    Usage: python evaluate_clustering.py graph.smat graph.clustering [outfile]
    """

class FastSet:
    """ A set over a finite universe of items. """
    def __init__(self, set, n):
        self.n = n
        self.member = [ False for _ in xrange(n) ]
        for i in set:
            self.member = True
    def __contains__(self,item):
        return self.member[item]
    
        
    
class Graph:
    def __init__(self, n, adj):
        self.adj = adj
        self.n = n
        nnz = 0
        for i in xrange(0,n):
            nnz += len(adj[i])
        self.nnz = nnz
        
    def volume(self,S):
        vol = 0
        for s in S:
            vol += len(self.adj[s])
        return vol
    
    def cutsize(self,S):
        cut = 0
        # index the set
        curset = set(S)
        
        for s in S:
            for j in self.adj[s]:
                if j not in curset:
                    cut += 1
        return cut
        

def read_smat(filename):
    file = open(filename,'rt')
    header = file.readline()
    parts = header.split()
    m = int(parts[0])
    n = int(parts[1])
    nz = int(parts[2])
    adj = [ [] for _ in xrange(0,m) ]
    for line in file:
        parts = line.split()
        i = int(parts[0])
        j = int(parts[1])
        adj[i].append(j)
    return Graph(m, adj)
    
class Clusters:
    def __init__(self, m, vertices):
        self.clusters = vertices
        self.nclusters = len(vertices)
        
        if self.nclusters != m:
            sys.stderr.write('Error wrong number of clusters\n')
            sys.exit(-1)
            
        maxvert = 0
        for i in xrange(self.nclusters):
            for j in self.clusters[i]:
                maxvert = max(maxvert, j)
        
        self.nverts=maxvert+1
        
    def overlap(self,i):
        """ Compute the number of vertices shared with other clusters 
        
        @param i the cluster id
        @return (nsets,noverlap) nsets - the number of clusters with
          overlapped vertices, noverlap - the number of my vertices
          overlapped in other clusters
        
        """
        cluster = self.clusters[i]
        cset = [False for _ in xrange(self.nverts)]
        for i in cluster:
            cset[i] = True
        
        oset = [False for _ in xrange(self.nverts)]    
            
        nsets = 0
        noverlap = 0
        for ci in xrange(self.nclusters):
            if ci==i: continue
            for j in self.clusters[ci]:
                anyoverlap = False
                if cset[j]:
                    anyoverlap = True
                    if oset[j] is False:
                        noverlap += 1
                        # we don't want to double-count vertices
                        oset[j] = True
                if anyoverlap:
                    nsets += 1
        
        return (nsets, noverlap)
                    
                
            
        
        
        
    def report(self, graph):
        if self.nverts != graph.n:
            sys.stderr.write('Error, wrong number of vertices %i != %i\n',
                self.nverts, graph.n)
            sys.exit(-1)

        for ci in xrange(self.nclusters):
            """
            report
            
            the size 
            the volume
            the cutsize
            the conductance
            # (not currently output)
            #the total number of intersecting clusters
            #the overlap with other clusters
            
            """
            size = len(self.clusters[ci])
            volume = graph.volume(self.clusters[ci])
            cutsize = graph.cutsize(self.clusters[ci])
            conductance = float(cutsize)/float(min(volume,graph.nnz-volume))
            #(nset,noverlap) = self.overlap(ci)
            
            
            sys.stdout.write('%i %i %i %f\n'%(
                size,volume,cutsize,conductance))
                   
       
        
    
    
        
def read_clusters(filename):
    file = open(filename,'rt') 
    header = file.readline()
    parts = header.split()
    nclusters = int(parts[0])
    nverts = int(parts[2])
    clusters = [ [] for _ in xrange(nclusters) ]

    for line in file:
        parts = line.split()
        ci = int(parts[0]) # cluster
        vi = int(parts[1]) # vertex
        clusters[ci].append(vi)
    return Clusters(nclusters, clusters)
    
           
    

def main():
    if len(sys.argv) < 3 or len(sys.argv) > 4:
        print usage()
        sys.exit(-1)
    
    graphfilename = sys.argv[1]
    clusterfilename = sys.argv[2]

    graph = read_smat(graphfilename)
    clusters = read_clusters(clusterfilename)
    
    if len(sys.argv) == 4:
        outfilename = sys.argv[3]
        sys.stdout = open(outfilename,'wt')
    
    clusters.report(graph)
    

if __name__=='__main__':
    main()
