import sys
import os
import time
import random
import subprocess
from math import ceil, floor
from optparse import OptionParser

class FastSet:
    """ A set over a finite universe of items. """
    def __init__(self, set, n):
        self.n = n
        self.member = [ False for _ in xrange(n) ]
        for i in set:
            self.member = True
    def __contains__(self,item):
        return self.member[item]
    
   
class Graph:
    def __init__(self, n, adj):
        self.adj = adj
        self.n = n
        nnz = 0
        for i in xrange(0,n):
            nnz += len(adj[i])
        self.nnz = nnz
        
    def nedges(self):
        return self.nnz
        
    def nverts(self):
        return self.n
        
    def volume(self,S):
        vol = 0
        for s in S:
            vol += len(self.adj[s])
        return vol
    
    def cutsize(self,S):
        cut = 0
        # index the set
        curset = set(S)
        
        for s in S:
            for j in self.adj[s]:
                if j not in curset:
                    cut += 1
        return cut
        
    def boundary(self,S):
        # index the set
        curset = set(S)
        cutset = set()
        for s in S:
            for j in self.adj[s]:
                if j not in curset:
                    cutset.add(j)
                    
        return cutset
        

def read_graph(filename):
    file = open(filename,'rt')
    header = file.readline()
    parts = header.split()
    m = int(parts[0])
    adj = []
    for line in file:
        parts = line.split()
        verts = []
        for p in parts:
            v = int(p)-1
            assert(v < m)
            assert(v >= 0)
            verts.append(v)
        adj.append(verts)
    while len(adj) < m:
        adj.append([])
    return Graph(m,adj)
        
  
def run_command(cmd):
    """ 
    Return all the output from a command.
    
    Input
    -----
    @param cmd the command line
    
    Return
    ------
    (code,out,err) are the 
      return code
      stdout stream
      stderr stream
    respectively.  
    
    """
    #parts = shlex.split(cmd)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, 
                            stderr=subprocess.PIPE, shell=True)
    (out,err) = p.communicate()
    rval = p.returncode
    return (rval,out,err)
    
def index_clusters(clusters):    
  nclusters = max([vc[1] for vc in clusters])+1
  cluster_vertices = [ [] for _ in xrange(nclusters) ]
  for vc in clusters:
      vertex = vc[0]
      cluster = vc[1]
      cluster_vertices[cluster].append(vertex)
  
  return cluster_vertices
  
def maximum_cluster_volume(graph,clusters):
  """ find the maximum volume of a cluster in a graph. """
  # we first have to index the clusters
  cluster_vertices = index_clusters(clusters)
 
  volumes = [ graph.volume(cluster) for cluster in cluster_vertices ]
  return max(volumes)
  
  
  
def pack_clusters(graph,clusters,maxvol):
  # transform into indexed form
  clusters = index_clusters(clusters)
  iclusters = set( i for i in xrange(len(clusters)) )
  
  cluster_volumes = [ graph.volume(c) for c in clusters ]
  
  # index the vertices
  vertex2cluster = [ -1 for _ in xrange(graph.nverts()) ]
  for i,c in enumerate(clusters):
      for v in c:
          vertex2cluster[v] = i

  
  newclusters = []
  curcluster = 0
  
  while len(iclusters) > 0:
      
      newcluster_index = iclusters.pop()
      newcluster = clusters[newcluster_index]
      
      curvol = graph.volume(newcluster)
      
      if curvol == 0:
          continue
      
      if curvol < maxvol:
          
          # merge in other clusters
          while len(iclusters) > 0:
              boundary = graph.boundary(newcluster)
              
              # look at clusters in order of boundary overlap
              # make sure the other clusters aren't taken and
              # won't exceed maxvol if we combine
              cluster_scores = {}
              max_score = -1
              max_cluster = -1
              for v in boundary:
                  vc = vertex2cluster[v]
                  assert(vc >= 0)
                  if cluster_volumes[vc] + curvol < maxvol and vc in iclusters:
                      cluster_scores[vc] = cluster_scores.get(vc,0) + 1
                      if cluster_scores[vc] > max_score:
                          max_score = cluster_scores[vc]
                          max_cluster = vc
            
              if max_cluster < 0:
                  # no clusters to add
                  break
                  
              # combine with vc
              assert(max_cluster in iclusters)
              newcluster.extend(clusters[max_cluster])
              curvol += cluster_volumes[max_cluster]
              iclusters.remove(max_cluster)
              
      # output this cluster
      for i in newcluster:
          newclusters.append((i,curcluster))
      curcluster += 1
  
  return (curcluster, newclusters)
