/** @file distributed_pagerank_simulation.hpp
 * Simulate computing PageRank on a set of overlapping clustersing.
 */

/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, 2010
 * 
 * History
 * -------
 * :2010-10-05: Initial coding
*/

#define NOMINMAX
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <algorithm>
#include <vector>
#include <limits>
#include <set>
#include <map>
#ifdef __APPLE__
#include <tr1/unordered_set>
#include <tr1/unordered_map>
#else 
#include <unordered_set>
#include <unordered_map>
#endif
#include <limits>

#if defined(_WIN32) || defined(_WIN64) || defined(__APPLE__)
#define tr1ns std::tr1
#else
#define tr1ns std
#endif
#include <queue>

#include "sparfun.h"
#include "sparvec.h" 
#include "vecfun.h"
#include "sparfun_util.h"


class cluster_info {
public:  
  tr1ns::unordered_map<int,double> local_residual;
  tr1ns::unordered_map<int,double> remote_residual;
  tr1ns::unordered_map<int,double> local_pagerank;
};


/** Compute PageRank within a cluster
 * The algorithm we use within each cluster is a push-style Richardson 
 * method.  This method assumes we have out-neighbor access to the graph.
 * 
 * @param r the graph
 * @param c the cluster
 * @param cinfo current state of PageRank vector within the cluster
 * @param alpha the PageRank value of alpha
 * @param tol the solution tolerance inside the cluster
 * @param maxit the maximum number of iterations to do.
 * @return the total amount of computation
 */

template <typename ClusterSet>
size_t compute_pagerank_in_cluster(sparserow* r, const ClusterSet& c,
          cluster_info& cinfo, double alpha, double tol, size_t maxit)
{
  size_t total_computation = 0;
  
  // run through the residual and add any nodes with a value exceeding 
  // the tolerance
  //std::unordered_set<int> queue_members;
  std::queue<int> queue;
  
  typedef tr1ns::unordered_map<int,double>::const_iterator const_map_iterator;
  
  double queue_tol = tol/(double)c.size();
  
  double sum_residual = 0;
  for (const_map_iterator it=cinfo.local_residual.begin(),
         itend=cinfo.local_residual.end(); it != itend; ++it) {
    if (it->second > queue_tol) {
      queue.push(it->first);
      //queue_members.insert(it->first);
      sum_residual += it->second;
    }
  }
  
  while (queue.size() > 0 && sum_residual>tol) {
    int v = queue.front();
    queue.pop();
    //queue_members.erase(v);
    
    // we are about to do this much work 
    total_computation += sr_degree(r,v);
    
    double topush = cinfo.local_residual[v];
    cinfo.local_residual[v] = 0.;
    cinfo.local_pagerank[v] += topush;
    sum_residual -= topush;
    
    double invdeg = 1./(double)sr_degree(r, v);
    double topush_neigh = alpha*topush*invdeg;
    
    for (int nzi=r->ai[v]; nzi<r->ai[v+1]; ++nzi) {
      int j = r->aj[nzi];
      if (c.count(j) > 0) {
        
        double old_resid = cinfo.local_residual[j];
        cinfo.local_residual[j] += topush_neigh;
        
        if (old_resid + topush_neigh > queue_tol && 
              old_resid <= queue_tol) {
          queue.push(j);
          //queue_members.insert(j);
        } 
        sum_residual += topush_neigh;
      } else {
        cinfo.remote_residual[j] += topush_neigh;
      }
    }
  }

  assert(sum_residual <= tol);
  
  return total_computation;
}

/** Aggregate the remote residual from within a cluster
 * 
 * @param residual the global residual
 * @param cset the cluster set
 * @param cinfo the cluster info
 * @return the total communication
 */
template <typename Residual, typename ClusterSet>
size_t communicate_residual(Residual& r, const ClusterSet& cset, 
          cluster_info& cinfo, double tol)
{
  size_t total_communication = 0;
  
  typedef tr1ns::unordered_map<int,double>::iterator map_iterator;
  
  for (map_iterator it=cinfo.remote_residual.begin(), 
        itend=cinfo.remote_residual.end(); it!=itend; ++it) {
    if (it->second > tol) {
      total_communication += 1;
      r[it->first] += it->second;
      it->second = 0.;
    }
  }
  return total_communication;
}    

/** Aggregate the PageRank vector back from all the cluster 
 */
template <typename Vector>
void aggregate_pagerank(Vector& pr, cluster_info& cinfo)
{
  typedef tr1ns::unordered_map<int,double>::iterator map_iterator;
  
  for (map_iterator it=cinfo.local_pagerank.begin(), 
        itend=cinfo.local_pagerank.end(); it!=itend; ++it) {
    pr[it->first] += it->second;
  }
}

/** Distribute the global residual back to local clusters
 */      
template <typename GlobalResidual, class ClusterSetType, class CoreMap>
size_t distribute_residual(
    const GlobalResidual& residual,
    const std::vector<ClusterSetType>& clusters,
    std::vector<cluster_info>& cinfo,
    CoreMap& core_map)
{
  int n=(int)residual.size();
  // initialize the cluster information
  for (int i=0; i<n; ++i) {
    cinfo[core_map[i]].local_residual[i] += residual[i];
  }
}    
    
/** Simulate a distributed computation of PageRank.
 * 
 * @param hist a stream of doubles to record stats of the history
 *   hist.push_back( iter ); // the current stage/iteration
 *   hist.push_back( computation_volume ); // the amount of computation in this stage
 *   hist.push_back( communication_volume ); // the amount of communication in this stage
 *   hist.push_back( residual_sum ); // the current sum of the residuals
 * 
 * We use a historm stream to let users pick between saving the output
 * directly to disk (if it's an ostream) or to memory (if it's a vector).
 */

template <class ClusterSetType, class CoreMap, typename PageRankVector,
  typename HistoryInfo>
size_t distributed_pagerank_simulation(sparserow *r, 
    std::vector<ClusterSetType>& clusters,
    const CoreMap& core_map,
    double alpha, double tol, size_t maxit,
    PageRankVector& prvec, HistoryInfo& hist)
{
  size_t nclusters = clusters.size();
  size_t total_communication = 0, total_computation = 0; // global totals
  // initialize the cluster information
  std::vector< cluster_info > cinfo(nclusters);
  for (int i=0; i<r->n; ++i) {
    cinfo[core_map[i]].local_residual[i] = 1.;
  }
  // allocate a vector of memory for the global residual
  std::vector< double > global_residual(r->n);
  // 
  for (size_t iter=0; iter<maxit; ++iter) {
    size_t stage_computation = 0, stage_communication = 0;
    for (size_t c=0; c<nclusters; ++c) {
      // solve local cluster
      stage_computation += compute_pagerank_in_cluster(
                              r, clusters[c], cinfo[c], alpha, tol, maxit);
      stage_communication += communicate_residual(
                              global_residual, clusters[c], cinfo[c], 
                              tol/(double)nclusters);
    }
    // now redistribute the residual, this doesn't change
    // the communication because we "would" have sent everything
    // to the right cluster initially, it's just easier in the code
    // to skip that step
    distribute_residual(global_residual, clusters, cinfo, core_map);
    
    // zero out the residual
    double residual_sum = 0.;
    for (int i=0; i<r->n; ++i) {
      residual_sum += global_residual[i];
      global_residual[i] = 0.;
    }
    
    total_computation += stage_computation;
    total_communication += stage_communication;
    
    std::cout << iter << " " << stage_computation << " " << stage_communication << " " << residual_sum << std::endl;

    hist.push_back( (double)iter );
    hist.push_back( (double)stage_computation );
    hist.push_back( (double)stage_communication );
    hist.push_back( (double)residual_sum );

    
    if (residual_sum < tol) {
      break;
    }
    
  }
  
  // aggregate the pagerank vector
  assert(prvec.size() >= r->n);
  for (size_t c=0; c<nclusters; ++c) {
    aggregate_pagerank(prvec, cinfo[c]);
  }
  
}

