/** @file pagerank-communication.cc
 * Use an overlapping clustering to compute a PageRank vector.  In this 
 * experiment, we report on the total communication volume and 
 * computation.  
 */

/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, 2010
 * 
 * History
 * -------
 * :2010-10-05: Initial coding
*/

#define NOMINMAX
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <algorithm>
#include <vector>
#include <limits>
#include <set>
#include <map>
#include <iterator>

#ifdef __APPLE__
#include <tr1/unordered_set>
#else 
#include <unordered_set>
#endif
#include <limits>

#if defined(_WIN32) || defined(_WIN64) || defined(__APPLE__)
#define tr1ns std::tr1
#else
#define tr1ns std
#endif

#include "sparfun.h"
#include "sparvec.h" 
#include "vecfun.h"
#include "sparfun_util.h"

#include "distributed_pagerank_simulation.hpp"

void usage(FILE* s) {
  const char *usagestr =
    "pagerank-communication graphfile corefile outprfile outhistfile\n"
    "             [alpha] [tolerance] [maxit]\n"
    "All options are positional.  To get the default value, specify _\n"
    "e.g. pagreank-communication graph.smat graph.cores pagerank.hist _ 0.01\n"
    "changes tolerance to 0.01 without modifying any of the other options"
    "\n"
    "Simulates a distributed computational of PageRank and tracks the\n"
    "total computation cost, the total communication cost, and the convergence.\n"
    "Formally, the program solves (I-alpha Pbar)x = e.  The default tolerance\n"
    "is set to 1 because that gives an effective tolerance of 1/n if there aren't\n"
    "too many dangling nodes." 
    "\n"
    "The main output is a history file with one line per iteration over all clusters\n"
    "  [<iteration> <total computation volume> <total communication volume> <residual sum>]\n"
    "\n"
    "The other output is a PageRank file for verfication purposes.\n"
    "\n"
    "defaults:\n"
    " alpha     : 0.85\n"
    " ntrials   : 1\n"
    " maxit     : 10000\n";
  fprintf(s,"%s\n",usagestr);
  exit(-1);
}

double double_arg(const char *a,double d) {
  if (!a) { return d; }
  if (strcmp(a,"_")==0) {
    return d; 
  } else {
    return atof(a);
  }
}

/** Load an overlapping clustering into a vector of sets
 * This loads an smat file into a vector of sets, one set per
 * cluster
 * @param filename the name of the file
 * @param clusters an uninitialized structure for the clusters
 * @param core_map an uniniitalized structure for the core map from
 *   vertices to number of cores
 */
int load_overlapping_clusters(const char *filename,
    std::vector< tr1ns::unordered_set<int> >& clusters,
    std::vector<int>& core_map)
{
    FILE *f = fopen(filename, "rt");
    if (!f) {
        fprintf(stderr, "error opening %s\n", filename);
        return (-1);
    }

    int m, n, nz;
    if (fscanf(f, "%u %u %u", &m, &n, &nz)==3) {
        clusters.resize(m);
        core_map.resize(n);
        for (size_t ci=0; ci<n; ++ci) {
            core_map[ci] = -1;
        }
        int i, j, nzi = nz;
        double a;
        while (nzi-- > 0) {
            if (fscanf(f, "%u %u %lf", &i, &j, &a) != 3) {
                fprintf(stderr, "error reading %s\n", filename);
                return (-1);
            }
            if (i >= 0 && i < m) { 
                clusters[i].insert(j);
                if (a==1) {
                    if (core_map[j]!=-1) {
                        fprintf(stderr,
                            "duplicate core detected: line %i, vertex %u\n",
                            nz-nzi+1, j);
                        return (-3);
                    }
                    core_map[j]=i;
                }
            } else { 
                fprintf(stderr, 
                    "vertex-cluster out of range (0,%i): line %i - %u %u %lf\n",
                    m, nz-nzi+1, i, j, a);
                return (-2);
            }
        }
    } else {
        fprintf(stderr, "error reading %s\n", filename);
        return (-1);
    }
    
    // check the core_map
    for (size_t i=0; i<n; i++) {
        if (core_map[i]==-1) {
            fprintf(stderr,
                "invalid core_map for %s, vertex %Zi unassigned\n", 
                filename, i);
                return (-11);
        }
    }
    
    return 0;
}
    

int main(int argc, char **argv) 
{
  using namespace std;
  double st;

  if (argc<5) { usage(stdout); } else if (argc>8) { usage(stderr); }
  const char *graphfilename = argv[1];
  const char *corefilename = argv[2];
  const char *outprfilename = argv[3];
  const char *outhistfilename = argv[4];
  double alpha=double_arg(argc>5?argv[5]:NULL,0.85);
  double tol=double_arg(argc>6?argv[6]:NULL,1.);
  size_t maxit = (size_t)double_arg(argc>7?argv[7]:NULL,10000);

      
  
  printf("Parameters\n");
  printf("----------\n");
  printf("%25s : %s\n", "graphfilename", graphfilename);
  printf("%25s : %s\n", "corefilename", corefilename);
  printf("%25s : %s\n", "outprfilename", outprfilename);
  printf("%25s : %s\n", "outhistfilename", outhistfilename);
  printf("%25s : %lf\n", "alpha", alpha);
  printf("%25s : %lg\n", "tolerance", tol);
  printf("%25s : %Zi\n", "maxit",maxit);
  printf("\n");

  sparserow Gs={0};
  sparserow Cs={0};
  int rval = 0;
  
  st = sf_time();
  rval |= load_generic(graphfilename,&Gs,false);
  if (rval) {
    cerr << "*** Error: couldn't open " << graphfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << graphfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    nodes = " << Gs.m << endl;
  cout << "    edges = " << sr_nedges(&Gs) << endl;
  cout << endl;
  
  st = sf_time();
  std::vector< tr1ns::unordered_set<int> > clusters;
  std::vector< int > core_map;
  rval |= load_overlapping_clusters(corefilename,clusters,core_map);
  if (rval) {
    cerr << "*** Error: couldn't open " << corefilename << endl;
    return (-10); 
  }
  cout << "Loaded " << corefilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    clusters = " << clusters.size() << endl;
  cout << endl;

  
  int nverts = Gs.n;
  assert(nverts == core_map.size());
  int nclusters = clusters.size();
  
  std::vector<double> prvec(Gs.n,0.);
  std::vector<double> history;
  
  distributed_pagerank_simulation(&Gs,
      clusters, core_map, alpha, tol, maxit,
      prvec, history);
      
    
      
  FILE *outfile = fopen(outhistfilename,"wt");
  if (outfile) {
    double total_communication = 0;
    double total_computation = 0;
    for (size_t i=0; i<history.size(); ) {
      fprintf(outfile, "%g ", history[i]); i++;
      fprintf(outfile, "%g ", history[i]); total_computation += history[i]; i++; 
      fprintf(outfile, "%g ", history[i]); total_communication += history[i]; i++;
      fprintf(outfile, "%g ", history[i]); i++;
      fprintf(outfile, "\n"); 
    }    
    fclose(outfile);
    cout << "total computation: " << total_computation << endl;
    cout << "total communication: " << total_communication << endl;
  } else {
    cerr << "*** Error: Couldn't write output to " << outhistfilename << endl;
    return (-1);
  }
  
  FILE *outprfile = fopen(outprfilename, "wt");
  if (outprfile) {
    for (size_t i=0; i<Gs.n; ++i) {
      fprintf(outprfile, "%18.16e\n", prvec[i]);
    }
    fclose(outprfile);
  } else {
    cerr << "*** Error: Couldn't write output to " << outprfilename << endl;
    return -1;
  }
  
  
  return 0;
}

