/** @file pagerank-escprobs.cc
 * Given a set of clusters on a graph, this code finds the cores
 * of each cluster, in the sense that the core of the cluster has
 * a low escaping probability for a PageRank random walk.
 */


/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, University of British Columbia, 2010
 * 
 * History
 * -------
 * :2010-02-05: Moved codes out of pagerank-containment.cc
 * 
 * Todos
 * -----
 */

#include <math.h>
#include <string.h>

#include "containment.h"
#include "vecfun.h"
 
/**
 * The expected exit time for a Markov chain from a cluster can
 * be computed with a simple application of first-transition analysis.
 * Let u(x) = E[T | X0=x] and T=min(n >= 0 : Xn is not in C)
 * Then, 
 *   u(x) = 0 if x is not in C
 *   u(x) = 1 + sum_y P(x,y) u(y) if x is in C
 * but u(y) = 0 if y is not in C, thus, the second expression simplifies
 * to u(x) = 1 + sum_{y in C} P(x,y) u(y).
 * Let B be the sub-matrix of P corresponding to nodes in C, then
 * u = e + B*u
 * must hold at the steady-state values of u, 
 * so we can just use a Neumann approximation 
 *
 * We'll just use a Neumann series to solve this system.
 * (which ignores some technical details about the process,
 * but c'est la vie.)  To wit, 
 *  u \approx u_k = sum(B^n e, n=0...k)
 * and we have u_k = B*u_{k-1} + e
 *
 * Recall that we're using a PageRank markov chain, so we have
 * to explicitly model the jump with alpha.  But, it'll make 
 * everything converge fast.
 *
 * group is an array of the elements of a group
 * groupsize is the size of the group array
 * iwork is an integer array of size G->m, 
 *   every element should=-1 on initialization
 * dwork is a double array of size groupsize
 * leavetimes is an array of groupsize and will contain
 *   the leavetime for each element in group
 * i.e. probs[i] is the leavetime for group[i]
 */
void compute_leavetimes(
      sparserow *G, int *group, int groupsize, 
      double alpha, double tol, int maxiter,
      int *iwork, double *dwork,
      double *leavetimes)
{
  // build iwork to map from the cluster to the local area
  for (int i = 0; i<groupsize; i++) { iwork[group[i]] = i; }
  // initialize x to f
  double *x = dwork, *y = leavetimes;
  for (int i = 0; i<groupsize; i++) {
    x[i] = 1;
  }
  int iter=0;
  while (iter++<=maxiter) {
    // since u_k=x compute y=B*u_k
    // compute the sum of x, the current iterate
    double sumx = csum(x,groupsize); // TODO optimize and put this sum in place
    double delta = 0;
    for (int i = 0; i<groupsize; i++) {
      int v = group[i];
      double yi=0.0;
      int nneighout=0;
      for (int nzi=G->ai[v]; nzi<G->ai[v+1]; ++nzi) {
        int ji = iwork[G->aj[nzi]]; // local index of neighbor
        if (ji>=0) { yi += x[ji]; } // vertex in the cluster
        else { nneighout++; } // vertex not in the cluster
      }
      // set y[i] = (B*u_k)_i
      y[i] = 0.;
      if (sr_degree(G,v)>0) {
        double invn=1.0/sr_degree(G,v);
        y[i] = alpha*invn*yi;
      }
      y[i] += (1-alpha)*sumx/(double)G->m;
      // now add e
      y[i] += 1;
      delta += fabs(y[i]-x[i]);
    }
    {double *temp = x; x = y; y = temp; }
#if _DEBUG
    printf("leavetime size=%5i iter=%2i : delta=%10.3e sumx=%8.2e \n", 
      groupsize, iter, delta, sumx);
#endif
    if (delta<tol) { break; }
  }
  // correct our flip-flopping!
  if (x != leavetimes) { memcpy(leavetimes, x, groupsize*sizeof(double)); }
#if _DEBUG
  for (int i = 0; i<groupsize; i++) { assert (x[i]>1); }
#endif
  // clear our work array
  for (int i = 0; i<groupsize; i++) { iwork[group[i]] = -1; }
}

/**
 * The escaping probability for a Markov chain is just an 
 * instance of first transition analysis.  Let u(x) be the
 * escaping probability for node x, then 
 *  u(x) = sum(P(x,y),y not in C) + sum(P(x,y)*u(y),y in C)
 * which yields the linear system
 *  u = f + B*u (B = P(C,C), or just the transitions between states in C)
 * over only the vertices in C. (P is the transition matrix
 * so P(x,y) is the transition prob between state x and y.)
 *
 * We'll just use a Neumann series to solve this system.
 * (which ignores some technical details about the process,
 * but c'est la vie.)  To wit, 
 *  u \approx u_k = sum(B^n f, n=0...k)
 * and we have u_k = B*u_{k-1} + f
 *
 * Recall that we're using a PageRank markov chain, so we have
 * to explicitly model the jump with alpha.  But, it'll make 
 * everything converge fast.
 *
 * group is an array of the elements of a group
 * groupsize is the size of the group array
 * iwork is an integer array of size G->m, 
 *   every element should=-1 on initialization
 * dwork is a double array of size groupsize
 * probs is an array of groupsize and will contain
 *   the escaping probability for each element in group
 * i.e. probs[i] is the escaping probability for group[i]
 */
void compute_escaping_probabilities(
      sparserow *G, int *group, int groupsize, 
      double alpha, double tol, int maxiter,
      int *iwork, double *dwork,
      double *probs)
{
  // build iwork to map from the cluster to the local area
  for (int i = 0; i<groupsize; i++) { iwork[group[i]] = i; }
  // initialize x to f
  double *x = dwork, *y = probs;
  for (int i = 0; i<groupsize; i++) {
    int v = group[i];
    int nneighout=0;
    for (int nzi=G->ai[v]; nzi<G->ai[v+1]; ++nzi) {
      if (iwork[G->aj[nzi]]<0) { nneighout++; } // neighbor not in the cluster
    }
    double xi = 0.0;
    if (sr_degree(G,v)>0) {
      xi += alpha*(double)nneighout/(double)sr_degree(G,v);
    }
    xi += (1-alpha)*(double)(G->m-groupsize)/(double)G->m;
    x[i] = xi;
  }
  int iter=0;
  while (iter++<=maxiter) {
    // since u_k=x compute y=B*u_k
    // compute the sum of x, the current iterate
    double sumx = csum(x,groupsize); // TODO optimize and put this sum in place
    double delta = 0;
    for (int i = 0; i<groupsize; i++) {
      int v = group[i];
      double yi=0.0;
      double invn=1.0/sr_degree(G,v);
      int nneighout=0;
      for (int nzi=G->ai[v]; nzi<G->ai[v+1]; ++nzi) {
        int ji = iwork[G->aj[nzi]]; // local index of neighbor
        if (ji>=0) { yi += x[ji]; } // vertex in the cluster
        else { nneighout++; } // vertex not in the cluster
      }
      // set y[i] = (B*u_k)_i
      y[i] = 0.;
      if (sr_degree(G,v)>0) {
        y[i] = alpha*invn*yi;
        // now add f
        y[i] += alpha*invn*(double)nneighout;
      }
      y[i] += (1-alpha)*sumx/(double)G->m;
      y[i] += (1-alpha)*(double)(G->m-groupsize)/(double)G->m;
      delta += fabs(y[i]-x[i]);
    }
    {double *temp = x; x = y; y = temp; }
#if _DEBUG
    printf("escprob size=%5i iter=%2i : delta=%10.3e sumx=%8.2e \n", 
      groupsize, iter, delta, sumx);
#endif
    if (delta<tol) { break; }
  }
  // correct our flip-flopping!
  if (x != probs) { memcpy(probs, x, groupsize*sizeof(double)); }
#if _DEBUG
  for (int i = 0; i<groupsize; i++) { assert (x[i]<1.000000000000001); }
#endif
  // clear our work array
  for (int i = 0; i<groupsize; i++) { iwork[group[i]] = -1; }
}
