/** @file pagerank-escprobs.cc
 * Given a set of clusters on a graph, this code finds the cores
 * of each cluster, in the sense that the core of the cluster has
 * a low escaping probability for a PageRank random walk.
 */


/**
 * Authors
 * -------
 * David F. Gleich
 * Copyright, University of British Columbia, 2010
 * 
 * History
 * -------
 * :2008-09: Initial coding
 * :2010-01-26: Revision and output
 * :2010-01-27: Added leavetimes, suggested name change to pagerank-containment
 * :2010-02-05: Moved the containment calculation to another cc file,
 *   so we can include it in other codes.  It's really fast, after all.
 * 
 * Todos
 * -----
*/

#define NOMINMAX
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <algorithm>
#include <vector>
#include <limits>

#include "sparfun_util.h"
#include "containment.h"

void usage(FILE* s) {
  const char *usagestr =
    "pagerank-containment graphfile hclusterfile containfile \n"
    "   [method] [pralpha] [maxsteps]\n"
    "All options are positional.  To get the default value, specify _\n"
    "e.g. pagerank-containment graph.smat graph.hcluster graph.escprobs _ 10\n"
    "uses the default pagerank alpha value, but changes the volpenalty to 10.\n"
    "\n"
    "Valid methods are either leavetime or stayprob\n'"
    "\n"
    "defaults:\n"
    " method        : leavetime\n"
    " pralpha       : 0.85\n"
    " maxsteps      : 25\n";
  fprintf(s,"%s\n",usagestr);
  exit(-1);
}

double double_arg(const char *a,double d) {
  if (!a) { return d; }
  if (strcmp(a,"_")==0) {
    return d; 
  } else {
    return atof(a);
  }
}

struct group_stats_t {
  int nisoverts; // number of vertices outside of all groups
  int maxgroups; // maximum times a vertex appears in any group
  double avggroups; // average times a vertex appears in any group
  int maxgroupsize; // largest size of any group
};


void group_stats(int nverts, sparserow *C, const char *clusterfilename,
                 group_stats_t* s = NULL) 
{
  printf("\n");
  printf("Cluster stats\n");
  printf("-------------\n");
  printf("%25s : %i\n", "number of groups", C->m);
  group_stats_t stats;
  if (!s) { s = &stats; } // use local storage if no structure
  s->nisoverts = 0;
  s->maxgroupsize = 0;
  s->maxgroups = 0;
  s->avggroups = 0.0;
  { 
    std::vector<int> count(nverts,0); 
    for (int i=0; i<C->m; ++i) {
      s->maxgroupsize = std::max(s->maxgroupsize,sr_degree(C,i));
      for (int nzi=C->ai[i]; nzi<C->ai[i+1]; ++nzi) { count[C->aj[nzi]]++; }
    } 
    for (int i=0; i<nverts; ++i) { 
      if (!count[i]) { s->nisoverts++; }
      else {
        s->maxgroups = std::max(s->maxgroups,count[i]);
        s->avggroups += count[i];
      }
    }
    s->avggroups /= (nverts-s->nisoverts);
  }
  printf("%25s : %i\n", "isolated vertices", s->nisoverts);
  printf("%25s : %i\n", "largest group", s->maxgroupsize);
  printf("%25s : %i\n", "max groups", s->maxgroups);
  printf("%25s : %lf\n", "average groups", s->avggroups);
  
  printf("\n");
}



/** Find the cluster with the best of the worse containments
 * That is, look at the vertex with worst containment in any cluster,
 * then find the cluster with the best of the worst.  This tells us
 * that a core had better be INSIDE the best of the worst containment.
 */
double compute_best_beta(sparserow* C)
{
  double maxval = std::numeric_limits<double>::max();
  std::vector<double> betas(C->n,maxval);
  for (int i=0; i<C->m; ++i) {
    for (int nzi=C->ai[i]; nzi<C->ai[i+1]; ++nzi) { 
      int v = C->aj[nzi];
      double contain = C->a[nzi]; // containment

      betas[C->aj[nzi]] = std::min(betas[C->aj[nzi]],contain);
    }
  }
  double bestbeta=std::numeric_limits<double>::min();
  for (int i=0; i<C->n; ++i) {
    if (betas[i]!=maxval) {
      bestbeta = std::max(betas[i],bestbeta);
    }
  }
  return bestbeta;
}

/** Assign vertices as cores of a cluster.
 * Given a compressed sparse row matrix where each row in a cluster
 * and the non-zeros in the row is the escaping probability of the
 * node in the cluster, this script assigns 
 * 
 * @param Cs the CSR matrix where each non-zero value is an escaping 
 *  probability
 * @param core_ind a vector mapping to the non-zero values in Cs, which
 *   is 1 if the vertex in the cluster is a core, or 0 otherwise.  This
 *   array has length nnz(Cs)
 * @param beta the value of beta to map a vertex to a core
 */
void assign_cores(const sparserow& Cs, int* core_ind, double beta)
{
    for (size_t i=0; i<Cs.ai[Cs.m]; i++) {
        if (Cs.a[i] < beta) {
            core_ind[i] = 1;
        } else { 
            core_ind[i] = 0;
        }
    }
}

int main(int argc, char **argv) 
{
  using namespace std;
  double st = 0;

  if (argc<4) { usage(stdout); } else if (argc>7) { usage(stderr); }
  const char *graphfilename = argv[1];
  const char *hclusterfilename = argv[2];
  const char *outfilename = argv[3];
  const char *method = argc>4?argv[4]:"leavetime";
  double alpha=double_arg(argc>5?argv[5]:NULL,0.85);
  int maxsteps=(int)double_arg(argc>6?argv[6]:NULL,25);


  printf("Parameters\n");
  printf("----------\n");
  printf("%25s : %s\n", "graphfilename", graphfilename);
  printf("%25s : %s\n", "hclusterfilename", hclusterfilename);
  printf("%25s : %s\n", "outfilename", outfilename);
  printf("%25s : %s\n", "method", method);
  printf("%25s : %f\n", "alpha", alpha);
  printf("%25s : %i\n", "maxsteps", maxsteps);
  printf("\n");

  sparserow Gs={0};
  sparserow Cs={0};
  int rval = 0;
  
  st = sf_time();
  rval |= load_generic(graphfilename,&Gs,false);
  if (rval) {
    cerr << "*** Error: couldn't open " << graphfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << graphfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    nodes = " << Gs.m << endl;
  cout << "    edges = " << sr_nedges(&Gs) << endl;
  cout << endl;
  
  st = sf_time();
  rval |= load_smat(hclusterfilename,&Cs,false);
  if (rval) {
    cerr << "*** Error: couldn't open " << hclusterfilename << endl;
    return (-10); 
  }
  cout << "Loaded " << hclusterfilename << " in " 
       << sf_time() - st << " secs" << endl;
  cout << "    clusters = " << Cs.m << endl;
  cout << "    assignments = " << sr_nedges(&Cs) << endl;
  cout << endl;
  
  
  int nverts = std::max(Gs.m,Cs.m);
  group_stats_t stats;
  group_stats(nverts,&Cs, hclusterfilename,&stats);

  //
  // compute the containment measure
  // we either have leavetime -- the number of steps until you leave
  //   the cluster 
  // or stayprob -- the probability of staying within the cluster
  // both of these values are high for elements in the core
  // of a cluster, but low for values on the edges.  
  //

  st = sf_time();
  // allocate space for all the probabilities
  size_t new_mem = sizeof(double)*Cs.ai[Cs.m];
  Cs.a = (double*)malloc(sizeof(double)*Cs.ai[Cs.m]);
  if (!Cs.a) { cerr << "*** Error: can't malloc space for escprobs" << endl; }
  {
    std::vector<int> iwork(Gs.m,-1);
    std::vector<double> dwork(stats.maxgroupsize);
    
    if (strcmp(method,"stayprob") == 0) {
      for (int gi = 0; gi<Cs.m; ++gi) {
#if _DEBUG
        cout << "Computing escape probabilities for cluster " << gi << endl;
#endif        
        compute_escaping_probabilities(&Gs,
          &Cs.aj[Cs.ai[gi]], Cs.ai[gi+1]-Cs.ai[gi], 
          alpha, 1e-4, maxsteps, &iwork[0], &dwork[0],
          &Cs.a[Cs.ai[gi]]);
        // need to switch to stay probabilities, so 1-escprob
        for (int nzi=Cs.ai[gi]; nzi<Cs.ai[gi+1]; ++nzi) {
          Cs.a[nzi] = 1. - Cs.a[nzi];
        }
      }
    } else if (strcmp(method,"leavetime") == 0) {
      for (int gi = 0; gi<Cs.m; ++gi) {
#if _DEBUG
        cout << "Computing leavetimes for cluster " << gi << endl;
#endif        
        compute_leavetimes(&Gs,
          &Cs.aj[Cs.ai[gi]], Cs.ai[gi+1]-Cs.ai[gi], 
          alpha, 1e-4, maxsteps, &iwork[0], &dwork[0],
          &Cs.a[Cs.ai[gi]]);
      }
    } else {
      cerr << "*** Error: unknown method " << method << endl;
      exit(-10);
    }
  }
  cout << "Computed " << method << " containment in " 
       << sf_time() - st << " secs" << endl;

  //
  // sort vertices
  //
  sf_sortvals(&Cs);

  // 
  // output information
  //
  st = sf_time();
  FILE *outfile = fopen(outfilename,"wt");
  if (outfile) {
    fprintf(outfile,"%u %u %u\n",Cs.m,Cs.n,Cs.ai[Cs.m]);
    for (int i = 0; i<Cs.m; ++i) {
        for (int ai=Cs.ai[i]; ai<Cs.ai[i+1]; ++ai) {
            fprintf(outfile, "%u %u %18.16e\n", i, Cs.aj[ai], Cs.a[ai]);
        }
    }
    fclose(outfile);
    cout << "Write " << outfilename << " in " 
         << sf_time() - st << " secs" << endl;
  } else {
    cerr << "*** Error: Couldn't write output to " << outfilename << endl;
    return (-1);
  }

  return 0;
}
